/*
 * Copyright (C) 2007 by Quentin Anciaux
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	@author Quentin Anciaux
 */
package org.allcolor.yahp.cl.converter;
import com.lowagie.text.pdf.PdfCopy;
import com.lowagie.text.pdf.PdfImportedPage;
import com.lowagie.text.pdf.PdfReader;
import com.lowagie.text.pdf.PdfSignatureAppearance;
import com.lowagie.text.pdf.PdfStamper;
import com.lowagie.text.pdf.PdfWriter;

import org.allcolor.yahp.converter.IHtmlToPdfTransformer;
import org.allcolor.yahp.converter.IHtmlToPdfTransformer.CConvertException;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.OutputStream;

import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.cert.Certificate;

import java.util.List;
import java.util.Map;


/**
 * Use iText to construct a complete pdf document from differents pdf
 * parts. Apply configured security policies on the resulting pdf.
 *
 * @author Quentin Anciaux
 * @version 0.94
 */
public class CDocumentReconstructor {
	/**
	 * construct a pdf document from pdf parts.
	 *
	 * @param files list containing the pdf to assemble
	 * @param properties converter properties
	 * @param fout outputstream to write the new pdf
	 * @param base_url base url of the document
	 * @param producer producer of the pdf
	 *
	 * @throws CConvertException if an error occured while reconstruct.
	 */
	public static void reconstruct(
		final List		   files,
		final Map		   properties,
		final OutputStream fout,
		final String	   base_url,
		final String	   producer)
		throws CConvertException {
		OutputStream out     = fout;
		boolean		 signed  = false;
		OutputStream oldOut  = null;
		File		 fsigned = null;

		try {
			fsigned = File.createTempFile("yahp", "pdf");

			if ("true".equals(properties.get(
							IHtmlToPdfTransformer.USE_PDF_SIGNING))) {
				signed     = true;
				oldOut     = out;
				out		   = new FileOutputStream(fsigned);
			} // end if

			com.lowagie.text.Document document = null;
			PdfCopy					  writer = null;
			boolean					  first  = true;

			for (int i = 0; i < files.size(); i++) {
				File	  fPDF   = (File) files.get(i);
				PdfReader reader = new PdfReader(fPDF.getAbsolutePath());
				reader.consolidateNamedDestinations();

				int n = reader.getNumberOfPages();

				if (first) {
					first = false;
					// step 1: creation of a document-object
					// set title/creator/author
					document     = new com.lowagie.text.Document(reader.getPageSizeWithRotation(
								1));

					// step 2: we create a writer that listens to the document
					writer		 = new PdfCopy(document, out);
					// compress the pdf
					writer.setFullCompression();
					// use pdf version 1.6
					writer.setPdfVersion(PdfWriter.VERSION_1_6);

					// check if encryption is needed
					if ("true".equals(properties.get(
									IHtmlToPdfTransformer.USE_PDF_ENCRYPTION))) {
						String password     = (String) properties.get(IHtmlToPdfTransformer.PDF_ENCRYPTION_PASSWORD);
						int    securityType = getSecurityFlags(properties);
						writer.setEncryption(PdfWriter.STRENGTH128BITS,
							password, null, securityType);
					} // end if

					String title = (String) properties.get(IHtmlToPdfTransformer.PDF_TITLE);

					if (title != null) {
						document.addTitle(title);
					} // end if
					else if (base_url != null) {
						document.addTitle(base_url);
					} // end else if

					String creator = (String) properties.get(IHtmlToPdfTransformer.PDF_CREATOR);

					if (creator != null) {
						document.addCreator(creator);
					} // end if
					else {
						document.addCreator(IHtmlToPdfTransformer.VERSION);
					} // end else

					String author = (String) properties.get(IHtmlToPdfTransformer.PDF_AUTHOR);

					if (author != null) {
						document.addAuthor(author);
					} // end if

					String sproducer = (String) properties.get(IHtmlToPdfTransformer.PDF_PRODUCER);

					if (sproducer != null) {
						document.addProducer(sproducer);
					} // end if
					else {
						document.addProducer(IHtmlToPdfTransformer.VERSION +
							" - http://www.allcolor.org/YaHPConverter/ - " +
							producer);
					} // end else

					// step 3: we open the document
					document.open();
				} // end if

				PdfImportedPage page;

				for (int j = 0; j < n;) {
					++j;
					page = writer.getImportedPage(reader, j);
					writer.addPage(page);
				} // end for
			} // end for

			document.close();
			out.flush();

			if (signed) {
				out.close();

				String    keypassword	   = (String) properties.get(IHtmlToPdfTransformer.PDF_SIGNING_PRIVATE_KEY_PASSWORD);
				String    password		   = (String) properties.get(IHtmlToPdfTransformer.PDF_ENCRYPTION_PASSWORD);
				String    keyStorepassword = (String) properties.get(IHtmlToPdfTransformer.PDF_SIGNING_KEYSTORE_PASSWORD);
				String    privateKeyFile   = (String) properties.get(IHtmlToPdfTransformer.PDF_SIGNING_PRIVATE_KEY_FILE);
				String    reason		   = (String) properties.get(IHtmlToPdfTransformer.PDF_SIGNING_REASON);
				String    location		   = (String) properties.get(IHtmlToPdfTransformer.PDF_SIGNING_LOCATION);
				boolean   selfSigned	   = !"false".equals((String) properties.get(
							IHtmlToPdfTransformer.USE_PDF_SELF_SIGNING));
				PdfReader reader		   = null;

				if (password != null) {
					reader = new PdfReader(fsigned.getAbsolutePath(),
							password.getBytes());
				} // end if
				else {
					reader = new PdfReader(fsigned.getAbsolutePath());
				} // end else

				KeyStore ks = selfSigned
					? KeyStore.getInstance(KeyStore.getDefaultType())
					: KeyStore.getInstance("pkcs12");
				ks.load(new FileInputStream(privateKeyFile),
					keyStorepassword.toCharArray());

				String	    alias   = (String) ks.aliases().nextElement();
				PrivateKey  key     = (PrivateKey) ks.getKey(alias,
						keypassword.toCharArray());
				Certificate chain[] = ks.getCertificateChain(alias);
				PdfStamper  stp     = PdfStamper.createSignature(reader,
						oldOut, '\0');

				if ("true".equals(properties.get(
								IHtmlToPdfTransformer.USE_PDF_ENCRYPTION))) {
					stp.setEncryption(PdfWriter.STRENGTH128BITS,
						password, null, getSecurityFlags(properties));
				} // end if

				PdfSignatureAppearance sap = stp.getSignatureAppearance();

				if (selfSigned) {
					sap.setCrypto(key, chain, null,
						PdfSignatureAppearance.SELF_SIGNED);
				} // end if
				else {
					sap.setCrypto(key, chain, null,
						PdfSignatureAppearance.WINCER_SIGNED);
				} // end else

				if (reason != null) {
					sap.setReason(reason);
				} // end if

				if (location != null) {
					sap.setLocation(location);
				} // end if

				stp.close();
				oldOut.flush();
			} // end if
		} // end try
		catch (final Exception e) {
			throw new CConvertException(
				"ERROR: An Exception occured while reconstructing the pdf document: " +
				e.getMessage(),
				e);
		} // end catch
		finally {
			try {
				fsigned.delete();
			} // end try
			catch (final Exception ignore) {}
		} // end finally
	} // end reconstruct()

	/**
	 * return the itext security flags for encryption
	 *
	 * @param properties the converter properties
	 *
	 * @return the itext security flags
	 */
	private static final int getSecurityFlags(final Map properties) {
		int securityType = 0;
		securityType     = "true".equals(properties.get(
					IHtmlToPdfTransformer.PDF_ALLOW_PRINTING))
			? (securityType | PdfWriter.AllowPrinting)
			: securityType;
		securityType     = "true".equals(properties.get(
					IHtmlToPdfTransformer.PDF_ALLOW_MODIFY_CONTENTS))
			? (securityType | PdfWriter.AllowModifyContents)
			: securityType;
		securityType     = "true".equals(properties.get(
					IHtmlToPdfTransformer.PDF_ALLOW_COPY))
			? (securityType | PdfWriter.AllowCopy)
			: securityType;
		securityType     = "true".equals(properties.get(
					IHtmlToPdfTransformer.PDF_ALLOW_MODIFT_ANNOTATIONS))
			? (securityType | PdfWriter.AllowModifyAnnotations)
			: securityType;
		securityType     = "true".equals(properties.get(
					IHtmlToPdfTransformer.PDF_ALLOW_FILLIN))
			? (securityType | PdfWriter.AllowFillIn)
			: securityType;
		securityType     = "true".equals(properties.get(
					IHtmlToPdfTransformer.PDF_ALLOW_SCREEN_READERS))
			? (securityType | PdfWriter.AllowScreenReaders)
			: securityType;
		securityType     = "true".equals(properties.get(
					IHtmlToPdfTransformer.PDF_ALLOW_ASSEMBLY))
			? (securityType | PdfWriter.AllowAssembly)
			: securityType;
		securityType     = "true".equals(properties.get(
					IHtmlToPdfTransformer.PDF_ALLOW_DEGRADED_PRINTING))
			? (securityType | PdfWriter.AllowDegradedPrinting)
			: securityType;

		return securityType;
	} // end getSecurityFlags()
} // end CDocumentReconstructor
