/*
 * Copyright (C) 2007 by Quentin Anciaux
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	@author Quentin Anciaux
 */
package org.allcolor.yahp.cl.converter.oo;
import com.sun.star.beans.XPropertySet;
import com.sun.star.bridge.XBridge;
import com.sun.star.bridge.XBridgeFactory;
import com.sun.star.comp.helper.Bootstrap;
import com.sun.star.connection.NoConnectException;
import com.sun.star.connection.XConnection;
import com.sun.star.connection.XConnector;
import com.sun.star.frame.XComponentLoader;
import com.sun.star.lang.DisposedException;
import com.sun.star.lang.EventObject;
import com.sun.star.lang.XComponent;
import com.sun.star.lang.XEventListener;
import com.sun.star.lang.XMultiComponentFactory;
import com.sun.star.uno.UnoRuntime;
import com.sun.star.uno.XComponentContext;

import org.apache.log4j.Logger;

import java.net.ConnectException;


/**
 * Represent an UNO connection to OpenOffice.org
 *
 * @author Quentin Anciaux
 * @version v0.94
  */
public class COOoConnection
	implements XEventListener {
	/** use for logging */
	private static Logger log = Logger.getLogger(COOoConnection.class);

	/** hostname where OOo is running on */
	private String hostName;

	/** OOo internal Api object */
	private XBridge bridge;

	/** OOo internal Api object */
	private XComponentContext context;

	/** OOo internal Api object */
	private XMultiComponentFactory serviceManager;

	/** if true reconnect automatically if the connection is closed */
	private boolean autoReconnecting = true;

	/** if true, the closing of the connection is expected to come */
	private boolean closingExpected = false;

	// bean properties
	/** port on which OOo is running on */
	private int port;

	/**
	 * Creates a new COOoConnection object.
	 */
	public COOoConnection() {
		this(8100, "localhost");
	} // end COOoConnection()

	/**
	 * Creates a new COOoConnection object.
	 *
	 * @param port port on which OOo is running on
	 */
	public COOoConnection(final int port) {
		this(port, "localhost");
	} // end COOoConnection()

	/**
	 * Creates a new COOoConnection object.
	 *
	 * @param port port on which OOo is running on
	 * @param hostName hostname where OOo is running on
	 */
	public COOoConnection(
		final int    port,
		final String hostName) {
		this.port		  = port;
		this.hostName     = hostName;
	} // end COOoConnection()

	/**
	 * set hostname where OOo is running on
	 *
	 * @param hostName The hostName to set.
	 */
	public final void setHostName(final String hostName) {
		this.hostName = hostName;
	} // end setHostName()

	/**
	 * @return Returns the hostName.
	 */
	public final String getHostName() {
		return hostName;
	} // end getHostName()

	/**
	 * set port on which OOo is running on
	 *
	 * @param port port on which OOo is running on
	 */
	public void setPort(final int port) {
		this.port = port;
	} // end setPort()

	/**
	 * @return Returns the port.
	 */
	public final int getPort() {
		return port;
	} // end getPort()

	/**
	 * Should we try to reestablish the connection if it goes down?
	 *
	 * @param autoReconnecting
	 */
	public void setAutoReconnecting(final boolean autoReconnecting) {
		this.autoReconnecting = autoReconnecting;
	} // end setAutoReconnecting()

	/**
	 * return the status of the auto reconnect flag
	 *
	 * @return the status of the auto reconnect flag
	 */
	public boolean getAutoReconnecting() {
		return autoReconnecting;
	} // end getAutoReconnecting()

	/**
	 * Obtains the com.sun.star.frame.Desktop service.
	 *
	 * @return com.sun.star.frame.Desktop service
	 *
	 * @throws ConnectException if an excetpion occurs
	 */
	public XComponentLoader getDesktop()
		throws ConnectException {
		return (XComponentLoader) UnoRuntime.queryInterface(XComponentLoader.class,
			getService("com.sun.star.frame.Desktop"));
	} // end getDesktop()

	/**
	 * Close the connection
	 */
	public void close() {
		closingExpected = true;
		((XComponent) UnoRuntime.queryInterface(XComponent.class, bridge)).dispose();
		closingExpected = false;
	} // end close()

	/**
	 * Connect to OOo
	 *
	 * @throws ConnectException if an error occurs
	 * @throws RuntimeException if an error occurs
	 */
	public void connect()
		throws ConnectException {
		log.info("connecting...");

		try {
			XComponentContext	   localContext		   = Bootstrap.createInitialComponentContext(null);
			XMultiComponentFactory localServiceManager = localContext.getServiceManager();
			XConnector			   connector		   = (XConnector) UnoRuntime.queryInterface(XConnector.class,
					localServiceManager.createInstanceWithContext(
						"com.sun.star.connection.Connector",
						localContext));
			XConnection			   connection		   = connector.connect(
					"socket,host=" + hostName + ",port=" + port);
			XBridgeFactory		   bridgeFactory	   = (XBridgeFactory) UnoRuntime.queryInterface(XBridgeFactory.class,
					localServiceManager.createInstanceWithContext(
						"com.sun.star.bridge.BridgeFactory",
						localContext));
			bridge = bridgeFactory.createBridge("", "urp", connection,
					null);
			((XComponent) UnoRuntime.queryInterface(XComponent.class,
				bridge)).addEventListener(this);
			serviceManager = (XMultiComponentFactory) UnoRuntime.queryInterface(XMultiComponentFactory.class,
					bridge.getInstance("StarOffice.ServiceManager"));

			XPropertySet set = (XPropertySet) UnoRuntime.queryInterface(XPropertySet.class,
					serviceManager);
			context = (XComponentContext) UnoRuntime.queryInterface(XComponentContext.class,
					set.getPropertyValue("DefaultContext"));
		} // end try
		catch (final Exception e) {
			log.error("connection failed " + e.getLocalizedMessage());

			if (e instanceof NoConnectException) {
				throw new ConnectException("connection failed; " +
					e.getMessage());
			} // end if

			throw new RuntimeException(e);
		} // end catch

		log.info("connected");
	} // end connect()

	/**
	 * sent by OOO when closing.
	 *
	 * @param event OOo event
	 */
	public void disposing(final EventObject event) {
		if (closingExpected) {
			log.info("closing");
		} // end if
		else {
			log.warn("closing unexpectedly");
		} // end else
	} // end disposing()

	/**
	 * Obtains a UNO service.
	 *
	 * @param service the UNO service name (e.g.
	 * 		  "com.sun.star.frame.Desktop")
	 *
	 * @return the service
	 *
	 * @throws ConnectException if the underlying connection went down
	 * @throws IllegalStateException if an error occurs
	 * @throws RuntimeException if an error occurs
	 */
	protected Object getService(final String service)
		throws ConnectException {
		if ((serviceManager == null) && !autoReconnecting) {
			throw new IllegalStateException("not connected");
		} // end if

		try {
			try {
				return serviceManager.createInstanceWithContext(service,
					context);
			} // end try
			catch (final DisposedException e) {
				log.warn("connection was closed");

				if (autoReconnecting) {
					connect();

					return serviceManager.createInstanceWithContext(service,
						context);
				} // end if
				else {
					throw new ConnectException("connection broken; " +
						e.getMessage());
				} // end else
			} // end catch
		} // end try
		catch (final Exception e) {
			if (e instanceof ConnectException) {
				throw (ConnectException) e;
			} // end if

			throw new RuntimeException("exception : serviceManager = " +
				serviceManager + " context = " + context, e);
		} // end catch
	} // end getService()
} // end COOoConnection
