/*
 * Copyright (C) 2007 by Quentin Anciaux
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	@author Quentin Anciaux
 */
package org.allcolor.yahp.cl.converter.oo;
import org.allcolor.yahp.cl.converter.CDocumentReconstructor;
import org.allcolor.yahp.converter.IHtmlToPdfTransformer;
import org.allcolor.yahp.converter.IHtmlToPdfTransformer.CConvertException;
import org.allcolor.yahp.converter.IHtmlToPdfTransformer.PageSize;

import org.apache.log4j.Logger;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;

import java.util.Arrays;
import java.util.Map;


/**
 * This class send an HTML document to be rendered in PDF by
 * OpenOffice.org. OpenOffice must be launched in Listen mode :
 * oowriter "-accept=socket,port=8100;urp;"
 *
 * @author Quentin Anciaux
 * @version 0.94
 */
public final class COOoHtmlRenderer {
	/** Use for logging */
	private static final Logger log = Logger.getLogger(COOoHtmlRenderer.class.getName());

	/**
	 * Constructor for the COOoHtmlRenderer object
	 */
	private COOoHtmlRenderer() {} // end COOoHtmlRenderer()

	/**
	 * Render the document in a pdf.
	 *
	 * @param size page size
	 * @param file document
	 * @param fbase_url base url of the document
	 * @param properties transform properties
	 * @param out outputstream to write into
	 *
	 * @throws CConvertException if an error occurs while rendering
	 */
	public static final void renderToPDF(
		final PageSize	   size,
		final String	   file,
		final String	   fbase_url,
		final Map		   properties,
		final OutputStream out)
		throws CConvertException {
		String base_url = fbase_url;
		base_url = base_url.substring(0, base_url.lastIndexOf("/"));

		final COOoHtmlRenderer g2r = new COOoHtmlRenderer();
		File				   fin = g2r.saveDocument(file, base_url);
		File				   f   = null;

		try {
			String host = ((properties != null) &&
					properties.containsKey(IHtmlToPdfTransformer.OOO_PDF_HOST))
				? (String) properties.get(IHtmlToPdfTransformer.OOO_PDF_HOST)
				: "localhost";
			String port = ((properties != null) &&
					properties.containsKey(IHtmlToPdfTransformer.OOO_PDF_PORT))
				? (String) properties.get(IHtmlToPdfTransformer.OOO_PDF_PORT)
				: "8100";
			COOoConverterFactory.setHostName(host);
			COOoConverterFactory.setPort(Integer.parseInt(port));

			try {
				f = File.createTempFile("yahp", ".pdf");
				f.delete();

				COOoConverter converter = COOoConverterFactory.getConverter();
				log.info("Convert : " + fin.toURI() + " into " +
					f.toURI());
				converter.convert(fin.toURI(), f.toURI());
				CDocumentReconstructor.reconstruct(Arrays.asList(new File[]{f}),properties,out,base_url,"OpenOffice.org Renderer");
			} // end try
			catch (final IOException e) {
				throw new CConvertException(
					"ERROR: An IOException occured while reconstructing the document: " +
					e.getMessage(),
					e);
			} // end catch
		} // end try
		finally {
			try {
				f.delete();
			} // end try
			catch (final Exception ignore) {}

			try {
				fin.delete();
			} // end try
			catch (final Exception ignore) {}
		} // end finally
	} // end renderToPDF()

	/**
	 * Sets the document attribute
	 *
	 * @param sfile The new document value
	 * @param base_url The base uri value
	 *
	 * @return DOCUMENT ME!
	 *
	 * @throws CConvertException DOCUMENT ME!
	 */
	private final File saveDocument(
		final String sfile,
		final String base_url)
		throws CConvertException {
		String file = sfile;
		file = file + "\n<script type=\"text/javascript\">\n" +
			"window.onload = function() { window.print();}\n" +
			"</script>\n";

		File tmp = null;

		try {
			tmp = File.createTempFile("yahp", ".html");

			FileOutputStream fo = new FileOutputStream(tmp);
			Writer			 w = new BufferedWriter(new OutputStreamWriter(
						fo,
						"utf-8"));
			w.write(file.toCharArray());
			w.flush();
			w.close();

			return tmp;
		} // end try
		catch (final IOException e) {
			try {
				tmp.delete();
			} // end try
			catch (final Exception ignore) {}

			throw new CConvertException(
				"ERROR: an IOException occured while saving document content: " +
				e.getMessage(),
				e);
		} // end catch
	} // end saveDocument()
} // end COOoHtmlRenderer
