/*
 * Copyright (C) 2007 by Quentin Anciaux
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	@author Quentin Anciaux
 */
package org.allcolor.yahp.cl.converter.swing;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.lang.ref.SoftReference;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JList;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.JTextPane;
import javax.swing.RepaintManager;
import javax.swing.SwingUtilities;
import javax.swing.event.DocumentEvent;
import javax.swing.text.AbstractDocument;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.Segment;
import javax.swing.text.View;
import javax.swing.text.ViewFactory;
import javax.swing.text.Position.Bias;
import javax.swing.text.html.CSS;
import javax.swing.text.html.FormView;
import javax.swing.text.html.HTMLDocument;
import javax.swing.text.html.HTMLEditorKit;
import javax.swing.text.html.ImageView;

import org.allcolor.css.parser.CStyler;
import org.allcolor.xml.parser.CShaniDomParser;
import org.allcolor.xml.parser.dom.ADocument;
import org.allcolor.yahp.cl.converter.CDocumentCut;
import org.allcolor.yahp.cl.converter.CDocumentReconstructor;
import org.allcolor.yahp.cl.converter.utils.CPadawan;
import org.allcolor.yahp.converter.CYaHPConverter;
import org.allcolor.yahp.converter.IHtmlToPdfTransformer;
import org.allcolor.yahp.converter.IHtmlToPdfTransformer.CConvertException;
import org.allcolor.yahp.converter.IHtmlToPdfTransformer.CHeaderFooter;
import org.allcolor.yahp.converter.IHtmlToPdfTransformer.PageSize;
import org.apache.fop.fonts.EmbedFontInfo;
import org.apache.fop.fonts.FontInfo;
import org.apache.fop.fonts.FontSetup;
import org.apache.fop.fonts.FontTriplet;
import org.apache.fop.fonts.apps.TTFReader;
import org.apache.fop.fonts.truetype.TTFFile;
import org.apache.fop.svg.PDFDocumentGraphics2D;
import org.apache.log4j.Logger;
import org.apache.xmlgraphics.java2d.GraphicContext;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.w3c.dom.css.CSSStyleDeclaration;


/**
 * This class render an HTML document in a pdf graphics 2D. It use a
 * JTextPane for html rendering.
 *
 * @author Quentin Anciaux
 * @version 0.98
 */
public final class CSwingHtmlRenderer {
	/** Use for logging */
	private static final Logger log = Logger.getLogger(CSwingHtmlRenderer.class.getName());

	private static final Map loadedFont = new Hashtable();
	
	/** The panel to render the html doc */
	private JTextPane panel;

	/**
	 * Constructor for the CWebRendererSwingHtmlRenderer object
	 */
	private CSwingHtmlRenderer() {
		panel = new JTextPane();
		panel.setEditorKitForContentType("text/html",
			new SynchronousEditorKit());
		panel.setEditable(false);
		panel.setDoubleBuffered(false);
		panel.setContentType("text/html");
		panel.setEditorKit(new SynchronousEditorKit());
	} // end CWebRendererSwingHtmlRenderer()

	private static CTextPanePrinter layoutDocument(Document doc,String base_url,int [] wh,int width) {
		CSwingHtmlRenderer g2r = new CSwingHtmlRenderer();
		g2r.setDocument(doc, base_url);
		g2r.layout(new Dimension(width, wh[1]));
		return new CTextPanePrinter(g2r.panel);
	}
	
	/**
	 * Render the document in a pdf.
	 *
	 * @param size page size
	 * @param node document
	 * @param fbase_url base url of the document
	 * @param hf headers footers list
	 * @param properties transform properties
	 * @param out outputstream to write into
	 *
	 * @throws CConvertException if an error occurs while rendering
	 */
	public static final void renderToPDF(
		final PageSize	   size,
		Document	   node,
		final String	   fbase_url,
		final List		   hf,
		final Map		   properties,
		final OutputStream out)
		throws CConvertException {
		String  base_url     = fbase_url;
		boolean textAsShapes = "true".equals(properties.get(
					IHtmlToPdfTransformer.HANDLE_CHAR_AS_GLYPH));
		String  encoding     = node.getInputEncoding();

		if (encoding != null) {
			if ((!encoding.toLowerCase().trim().equals("utf-8")) &&
					(encoding.toLowerCase().indexOf("iso") == -1)) {
				textAsShapes = true;

				if ("false".equals(properties.get(
								IHtmlToPdfTransformer.HANDLE_CHAR_AS_GLYPH))) {
					textAsShapes = false;
				} // end if
			} // end if
		} // end if

		base_url = base_url.substring(0, base_url.lastIndexOf("/"));

		final int isize[]   = size.getSize();
		final int imargin[] = size.getMargin();
		int		  width     = isize[0] - (imargin[0] + imargin[1]);
		int		  wh[]	    = size.getSize();
		List	  files     = new ArrayList();

		try {
			Document		 docs[]     = CDocumentCut.cut(node);
			node = null;
			SoftReference printers[] = new SoftReference[docs.length];
			int				 steps[]    = new int[docs.length];
			int				 totalpages = 0;
			log.info("Calculating nbr of pages... on "+docs.length+" document(s).");

			for (int i = 0; i < docs.length; i++) {
				NodeList bodies = docs[i].getElementsByTagName("body");
				if (bodies.getLength() > 0) {
					Element body = (Element)bodies.item(0);
					Node first = body.getFirstChild();
					Element table = docs[i].createElement("table");
					table.setAttribute("width", ""+width);
					table.setAttribute("height", "0");
					table.setAttribute("border", "0");
					if (first == null) {
						body.appendChild(table);
					} else {
						body.insertBefore(table, first);
					}
				}
				log.info("Layouting...");
				CTextPanePrinter printer = layoutDocument(docs[i],base_url,wh,width);
				printers[i] = new SoftReference(printer);
				log.info("Counting...");
				steps[i]	    = printer.countPages(size);
				totalpages += steps[i];
				log.info("nbr steps : "+steps[i]+" current total : "+totalpages);
			} // end for

			// have number of pages.
			log.info("Will render : " + totalpages + " page(s).");

			int numpage = 1;

			for (int k = 0; k < docs.length; k++) {
				CTextPanePrinter printer = (CTextPanePrinter)printers[k].get();
				if (printer == null) {
					log.info("Relayout, Softreference garbage collected.");
					printer = layoutDocument(docs[k],base_url,wh,width);
				}
				printers[k] = null;
				docs[k] = null;
				for (int i = 0; i < steps[k]; i++) {
					long tp1 = System.currentTimeMillis();
					log.info("begin pdf render page " + numpage + "/" +
						totalpages);

					try {
						File f = File.createTempFile("pdf", "yahp");
						files.add(f);

						FileOutputStream		    fout = new FileOutputStream(f);
						final PDFDocumentGraphics2D pdf = new PDFDocumentGraphics2D(textAsShapes,
								fout, size.getSize()[0],
								size.getSize()[1]);
						initFont(pdf.getFontInfo(),properties);
						pdf.initialize();

						GraphicContext gc = new GraphicContext();
						pdf.setGraphicContext(gc);
						pdf.setBackground(Color.WHITE);
						pdf.setColor(Color.WHITE);
						pdf.fillRect(0, 0, wh[0], wh[1]);

						if (SwingUtilities.isEventDispatchThread()) {
							int priority = Thread.currentThread()
													 .getPriority();

							try {
								Thread.currentThread()
										  .setPriority(Thread.MIN_PRIORITY);
								printer.print(pdf, size, i);
							} // end try
							finally {
								Thread.currentThread()
										  .setPriority(priority);
							} // end finally
						} // end if
						else {
							final int			   index = i;
							final CTextPanePrinter frd = printer;
							SwingUtilities.invokeAndWait(new Runnable() {
									public void run() {
										int priority = Thread.currentThread()
																 .getPriority();

										try {
											Thread.currentThread()
													  .setPriority(Thread.MIN_PRIORITY);
											frd.print(pdf, size, index);
										} // end try
										finally {
											Thread.currentThread()
													  .setPriority(priority);
										} // end finally
									} // end run()
								} // end new
							);
						} // end else
						renderHeaderFooter(base_url, hf, pdf, numpage,
							totalpages, size);
						pdf.finish();
						fout.close();
					} // end try
					catch (final Throwable ignore) {
						throw new CConvertException(
							"ERROR: error while painting into FOP PDFDocumentGraphics2D. Aborting.\n" +
							"page n°: " + (i + 1) + " base url: " +
							base_url,
							ignore);
					} // end catch

					long tp2 = System.currentTimeMillis();
					log.info("end pdf render page " + numpage + "/" +
						totalpages + " time : " + (tp2 - tp1) + "ms.");
					numpage++;
				} // end for
				printer = null;
			} // end for

			CDocumentReconstructor.reconstruct(files, properties, out,
				base_url, "Swing Renderer");
		} // end try
		finally {
			for (Iterator it = files.iterator(); it.hasNext();) {
				File f = (File) it.next();

				try {
					f.delete();
				} // end try
				catch (final Exception ignore) {}
			} // end for
		} // end finally
	} // end renderToPDF()

	private synchronized static void registerTTFFile(List fonts,File f) {
		if (f.isDirectory()) {
			File [] c = f.listFiles();
			if (c != null) {
				for (int i=0;i<c.length;i++) {
					registerTTFFile(fonts,c[i]);
				}
			}
		} else if (f.isFile() && f.getName().toLowerCase().endsWith(".ttf")) {
			try {
				File tmpFontMetric = null;
				Font font = null;
				String windowsName = null;
				if (loadedFont.containsKey(f.getAbsolutePath())) {
					Object [] obj = (Object [])loadedFont.get(f.getAbsolutePath());
					tmpFontMetric = (File)obj[0];
					font = (Font)obj[1];
					windowsName = (String)obj[2];
				} else {
					tmpFontMetric = File.createTempFile(f.getName(), ".xml");
					CYaHPConverter.registerFileToDeleteOnDestroy(tmpFontMetric);
					font = Font.createFont(Font.TRUETYPE_FONT, f.toURI().toURL().openStream());
					TTFReader app = new TTFReader();
	                TTFFile ttf = app.loadTTF(f.getAbsolutePath(), font.getFamily());
	                if (ttf != null) {
	                	windowsName = ttf.getWindowsName();
	                    org.w3c.dom.Document doc = app.constructFontXML(ttf,
	                            null, null, null, null, true,font.getFamily());
	                    if (doc != null) {
	                        app.writeFontXML(doc, tmpFontMetric.getAbsolutePath());
	    	                loadedFont.put(f.getAbsolutePath(), new Object[]{tmpFontMetric,font,windowsName});
	                    }
	                }
				}
				if (windowsName == null) return;
				List     triplets	   = new ArrayList();
				if (windowsName.toLowerCase().indexOf("bold") != -1 && windowsName.toLowerCase().indexOf("italic") != -1) {
					triplets.add(new FontTriplet(font.getFamily(),"italic",800));
				} else
				if (windowsName.toLowerCase().indexOf("italic") != -1) {
					triplets.add(new FontTriplet(font.getFamily(),"italic",400));
				} else
				if (windowsName.toLowerCase().indexOf("bold") != -1) {
					triplets.add(new FontTriplet(font.getFamily(),"normal",800));
				} else {
					triplets.add(new FontTriplet(font.getFamily(),"normal",400));
				}
				EmbedFontInfo ei = new EmbedFontInfo(tmpFontMetric.toURI().toURL().toExternalForm(),
						true, triplets, f.toURI().toURL().toExternalForm());
				fonts.add(ei);
			}
			catch (Throwable e){
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Init font embedding for fop
	 *
	 * @param info PDF fop font info object.
	 */
	private static void initFont(final FontInfo info,Map properties) {
		try {
			String pathToTTFFont = (String)properties.get(IHtmlToPdfTransformer.FOP_TTF_FONT_PATH);
			URL url = ClassLoader.getSystemResource(
					"yahp-fop-config.xml");

			if (url == null) {
				url = CSwingHtmlRenderer.class.getResource(
						"yahp-fop-config.xml");
			} // end if

			List		    fonts = new ArrayList();

			if (url != null) {
				CShaniDomParser p     = new CShaniDomParser();
				Document	    doc   = p.parse(url);
				NodeList	    nl    = doc.getElementsByTagName("font");

				for (int i = 0; i < nl.getLength(); i++) {
					Element  font		   = (Element) nl.item(i);
					String   xmlMetricPath = font.getAttribute(
							"metrics-file");
					boolean  kerning	   = "yes".equals(font.getAttribute(
								"kerning"));
					String   embedFontPath = font.getAttribute(
							"embed-file");
					List     triplets	   = new ArrayList();
					NodeList nli		   = font.getElementsByTagName(
							"font-triplet");

					for (int j = 0; j < nli.getLength(); j++) {
						Element     ft	    = (Element) nli.item(j);
						String	    name    = ft.getAttribute("name");
						String	    style   = ft.getAttribute("style");
						int	    weight  = "normal".equals(ft.getAttribute(
									"weight"))
							? 400
							: 800;
						FontTriplet triplet = new FontTriplet(name,
								style, weight);
						triplets.add(triplet);
					} // end for

					EmbedFontInfo ei = new EmbedFontInfo(xmlMetricPath,
							kerning, triplets, embedFontPath);
					fonts.add(ei);
				} // end for
			}
			if (pathToTTFFont != null) {
				File f = new File(pathToTTFFont);
				if (f.exists() && f.isDirectory()) {
					registerTTFFile(fonts,f);
				}
			}
			FontSetup.setup(info, fonts, null);
		} // end try
		catch (final Throwable e) {
			// ignore
		} // end catch
	} // end initFont()

	/**
	 * Sets the document attribute
	 *
	 * @param node The new document value
	 * @param base_url The base uri value
	 */
	private final void setDocument(
		final Document   node,
		final String base_url) {
		if (node.getNodeType() == Node.DOCUMENT_NODE) {
			ADocument doc = (ADocument)node;
			validateForSwingHtmlComponent(node, base_url,doc.getStyler());
		} else {
			ADocument doc = (ADocument)node.getOwnerDocument();
			validateForSwingHtmlComponent(node, base_url,doc.getStyler());
		}

		NodeList nl = ((Element) node).getElementsByTagName("link");

		if (nl.getLength() > 0) {
			List toRemove = new ArrayList();

			for (int i = 0; i < nl.getLength(); i++) {
				toRemove.add(nl.item(i));
			} // end for

			for (int i = 0; i < toRemove.size(); i++) {
				Node n = (Node) toRemove.get(i);
				n.getParentNode().removeChild(n);
			} // end for
		} // end if

		nl = node.getElementsByTagName("meta");

		if (nl.getLength() > 0) {
			List toRemove = new ArrayList();

			for (int i = 0; i < nl.getLength(); i++) {
				toRemove.add(nl.item(i));
			} // end for

			for (int i = 0; i < toRemove.size(); i++) {
				Node n = (Node) toRemove.get(i);
				n.getParentNode().removeChild(n);
			} // end for
		} // end if

		nl = node.getElementsByTagName("base");

		if (nl.getLength() == 0) {
			nl = node.getElementsByTagName("head");

			if (nl.getLength() == 1) {
				ADocument doc = node.getOwnerDocument() != null ? (ADocument)node.getOwnerDocument() : (ADocument)node;
				Element head = (Element) nl.item(0);
				Element base = doc.createElement("base");
				base.setAttribute("href", base_url);

				if (head.getFirstChild() != null) {
					head.insertBefore(base, head.getFirstChild());
				} // end if
				else {
					head.appendChild(base);
				} // end else
			} // end if
		} else {
			Element base = (Element) nl.item(0);
			base.setAttribute("href", base_url);
		}

		StringWriter w = new StringWriter();
		CPadawan.getInstance().rewriteXml(w, node, null);
		panel.setText(w.getBuffer().toString());
		((AbstractDocument) panel.getDocument()).setAsynchronousLoadPriority(-1);
	} // end setDocument()

	/**
	 * Return true if node is in the body section
	 *
	 * @param fnode node to check
	 *
	 * @return true if node is in the body section
	 */
	private final boolean isInBody(final Node fnode) {
		Node node = fnode;

		if (node.getNodeName().equals("body")) {
			return true;
		} // end if

		while (node.getParentNode() != null) {
			node = node.getParentNode();

			if (node.getNodeName().equals("body")) {
				return true;
			} // end if
		} // end while

		return false;
	} // end isInBody()

	/**
	 * return the size image needed to render the document without
	 * anything going off the side. could be different than the
	 * dimensions passed into layout because of text that couldn't
	 * break or a table that's set to be too big
	 *
	 * @param width page with
	 *
	 * @return The minimumSize value
	 */
	private final Rectangle getMinimumSize(final int width) {
		return new Rectangle(0, 0, (int) width,
			(int) panel.getPreferredSize().getHeight());
	} // end getMinimumSize()

	/**
	 * Layout the document
	 *
	 * @param dim page size
	 */
	private final void layout(final Dimension dim) {
		RepaintManager.currentManager(panel).markCompletelyDirty(panel);
		RepaintManager.currentManager(panel).validateInvalidComponents();
		panel.setSize(dim);
		panel.paintImmediately(getMinimumSize(dim.width));
	} // end layout()

	/**
	 * Render the header/footer on the current page
	 *
	 * @param urlForBase base url of the html document
	 * @param hf header-footer list
	 * @param g pdf graphics2d
	 * @param pagenum current page number
	 * @param pagetotal total page
	 * @param size pdf page size
	 */
	private static final void renderHeaderFooter(
		final String	 urlForBase,
		final List		 hf,
		final Graphics2D g,
		final int		 pagenum,
		final int		 pagetotal,
		final PageSize   size) {
		CShaniDomParser parser = new CShaniDomParser();

		if (hf.size() == 0) {
			return;
		} // end if

		final int     isize[]	   = size.getSize();
		final int     imargin[]    = size.getMargin();
		final int     headerHeight = imargin[3];
		final int     footerHeight = imargin[2];

		CHeaderFooter footer	   = null;

		for (int i = 0; i < hf.size(); i++) {
			CHeaderFooter e = (CHeaderFooter) hf.get(i);

			if ("footer".equals(e.getType())) {
				String sfor = e.getSfor();

				if ("odd".equals(sfor) && ((pagenum % 2) != 0)) {
					footer = e;

					break;
				} // end if

				if ("even".equals(sfor) && ((pagenum % 2) == 0)) {
					footer = e;

					break;
				} // end if

				if ("all".equals(sfor)) {
					if (footer == null) {
						footer = e;
					} // end if
				} // end if
			} // end if
		} // end for

		if (footer != null) {
			StringBuffer buffer = new StringBuffer();
			buffer.append("<html><head><title>empty</title></head>\n");
			buffer.append("<body><div style=\"width: ");
			buffer.append(isize[0]);
			buffer.append("px; height: ");
			buffer.append(headerHeight);
			buffer.append("px; \" >\n");
			buffer.append(footer.getContent());
			buffer.append("\n</div></body></html>");

			String res = buffer.toString();
			res     = res.replaceAll("\\$pagenum", "" + pagenum);
			res     = res.replaceAll("\\$pagetotal", "" + pagetotal);

			try {
				Document				 doc = parser.parse(new StringReader(
							res));
				final CSwingHtmlRenderer g2r = new CSwingHtmlRenderer();
				g2r.setDocument(doc, urlForBase);
				g2r.layout(new Dimension(isize[0], headerHeight));

				final PageSize psize = new PageSize(isize[0] -
						imargin[0], isize[1], 0, imargin[1], 0, 0);
				g.setTransform(new AffineTransform(4.0, 0.0, 0.0, 4.2,
						0.0, 0.0));
				g.translate(imargin[0], isize[1] - footerHeight);

				final CTextPanePrinter rd = new CTextPanePrinter(g2r.panel);

				if (SwingUtilities.isEventDispatchThread()) {
					int priority = Thread.currentThread().getPriority();

					try {
						Thread.currentThread()
								  .setPriority(Thread.MIN_PRIORITY);
						rd.printAll(g, psize);
					} // end try
					finally {
						Thread.currentThread().setPriority(priority);
					} // end finally
				} // end if
				else {
					SwingUtilities.invokeAndWait(new Runnable() {
							public void run() {
								int priority = Thread.currentThread()
														 .getPriority();

								try {
									Thread.currentThread()
											  .setPriority(Thread.MIN_PRIORITY);
									rd.printAll(g, psize);
								} // end try
								finally {
									Thread.currentThread()
											  .setPriority(priority);
								} // end finally
							} // end run()
						} // end new
					);
				} // end else
			} // end try
			catch (final Exception ignore) {}
		} // end if

		CHeaderFooter header = null;

		for (int i = 0; i < hf.size(); i++) {
			CHeaderFooter e = (CHeaderFooter) hf.get(i);

			if ("header".equals(e.getType())) {
				String sfor = e.getSfor();

				if ("odd".equals(sfor) && ((pagenum % 2) != 0)) {
					header = e;

					break;
				} // end if

				if ("even".equals(sfor) && ((pagenum % 2) == 0)) {
					header = e;

					break;
				} // end if

				if ("all".equals(sfor)) {
					if (header == null) {
						header = e;
					} // end if
				} // end if
			} // end if
		} // end for

		if (header != null) {
			StringBuffer buffer = new StringBuffer();
			buffer.append("<html><head><title>empty</title></head>\n");
			buffer.append("<body><div style=\"width: ");
			buffer.append(isize[0]);
			buffer.append("px; height: ");
			buffer.append(headerHeight);
			buffer.append("px; \" >\n");
			buffer.append(header.getContent());
			buffer.append("\n</div></body></html>");

			String res = buffer.toString();
			res     = res.replaceAll("(\\$pagenum)", "" + pagenum);
			res     = res.replaceAll("(\\$pagetotal)", "" + pagetotal);

			try {
				Document				 doc = parser.parse(new StringReader(
							res));
				final CSwingHtmlRenderer g2r = new CSwingHtmlRenderer();
				g2r.setDocument(doc, urlForBase);
				g2r.layout(new Dimension(isize[0], headerHeight));

				final PageSize psize = new PageSize(isize[0] -
						imargin[0], isize[1], 0, imargin[1], 0, 0);
				g.setTransform(new AffineTransform(4.0, 0.0, 0.0, 4.2,
						0.0, 0.0));
				g.translate(imargin[0], 0);

				final CTextPanePrinter rd = new CTextPanePrinter(g2r.panel);

				if (SwingUtilities.isEventDispatchThread()) {
					int priority = Thread.currentThread().getPriority();

					try {
						Thread.currentThread()
								  .setPriority(Thread.MIN_PRIORITY);
						rd.printAll(g, psize);
					} // end try
					finally {
						Thread.currentThread().setPriority(priority);
					} // end finally
				} // end if
				else {
					SwingUtilities.invokeAndWait(new Runnable() {
							public void run() {
								int priority = Thread.currentThread()
														 .getPriority();

								try {
									Thread.currentThread()
											  .setPriority(Thread.MIN_PRIORITY);
									rd.printAll(g, psize);
								} // end try
								finally {
									Thread.currentThread()
											  .setPriority(priority);
								} // end finally
							} // end run()
						} // end new
					);
				} // end else
			} // end try
			catch (final Exception ignore) {}
		} // end if

	} // end renderHeaderFooter()

	/**
	 * Remove problematic node, translate img uri to absolute. Getting
	 * style attribute for element node
	 *
	 * @param fnode current node
	 * @param fbase_url document base url
	 */
	private final void validateForSwingHtmlComponent(
		final Node   fnode,
		final String fbase_url,
		final CStyler styler) {
		Node   node     = fnode;
		String base_url = fbase_url;
		if ((node.getNodeType() == Node.DOCUMENT_TYPE_NODE)) {
			node.getParentNode().removeChild(node);

			return;
		} // end if
		else if ((node.getNodeType() == Node.COMMENT_NODE) &&
				isInBody(node)) {
			node.getParentNode().removeChild(node);

			return;
		} // end else if
		else if (node.getNodeName().equals("fieldset")) {
			NodeList nl = ((Element)node).getElementsByTagName("legend");
			String legend = null;
			if (nl.getLength() == 1) {
				legend = nl.item(0).getTextContent();
				nl.item(0).getParentNode().removeChild(nl.item(0));
				Element div = node.getOwnerDocument().createElement("div");
				Text t = node.getOwnerDocument().createTextNode(((char)160)+legend);
				div.appendChild(t);
				node.getParentNode().insertBefore(div, node);
			}
			Element div = node.getOwnerDocument().createElement("div");
			div.setAttribute("style", "border: 1px solid black;padding: 5px;");
			nl = node.getChildNodes();
			for (int i = 0; i < nl.getLength(); i++) {
				try{
					div.appendChild(nl.item(i).cloneNode(true));
				}
				catch (Exception ignore){}
			} // end for
			node.getParentNode().insertBefore(div, node);
			node.getParentNode().removeChild(node);
			node = div;
		} // end else if
		else if (node.getNodeName().equals("script")) {
			node.getParentNode().removeChild(node);

			return;
		} // end else if
		else if (node.getNodeName().equals("input")) {
			Element input = (Element) node;

			if ("hidden".equalsIgnoreCase(input.getAttribute("type"))) {
				node.getParentNode().removeChild(node);

				return;
			} // end if

			if ("button".equals(input.getAttribute("type"))) {
				input.setAttribute("type", "submit");
			} // end if
		} // end else if
		else if (node.getNodeName().equals("noscript")) {
			node.getParentNode().removeChild(node);

			return;
		} // end else if
		else if (node.getNodeName().equals("spacer")) {
			node.getParentNode().removeChild(node);

			return;
		} // end else if
		else if (node.getNodeName().equals("meta")) {
			node.getParentNode().removeChild(node);

			return;
		} // end else if
		else if (node.getNodeName().equals("label")) {
			NodeList nl = node.getChildNodes();

			for (int i = 0; i < nl.getLength(); i++) {
				try{
					node.getParentNode()
						.insertBefore(nl.item(i).cloneNode(true), node);
				}
				catch (Exception ignore){}
			} // end for

			node.getParentNode().removeChild(node);

			return;
		} // end else if
		else if (node.getNodeName().equals("acronym")) {
			NodeList nl = node.getChildNodes();

			for (int i = 0; i < nl.getLength(); i++) {
				node.getParentNode()
						.insertBefore(nl.item(i).cloneNode(true), node);
			} // end for

			node.getParentNode().removeChild(node);

			return;
		} // end else if
		else if (node.getNodeName().equals("a")) {
			Element newnode = node.getOwnerDocument()
									  .createElement("span");
			newnode.setAttribute("style",
				"color: #5b45ff;" +
				((Element) node).getAttribute("style"));
			newnode.setAttribute("class",
				((Element) node).getAttribute("class"));

			NodeList nl = node.getChildNodes();

			for (int i = 0; i < nl.getLength(); i++) {
				newnode.appendChild(nl.item(i).cloneNode(true));
			} // end for

			node.getParentNode().insertBefore(newnode, node);
			node.getParentNode().removeChild(node);
			node = newnode;
		} // end else if
		else if (node.getNodeName().equals("img")) {
			NamedNodeMap nnm	  = node.getAttributes();
			List		 toRemove = new ArrayList();

			for (int i = 0; i < nnm.getLength(); i++) {
				Attr a = (Attr) nnm.item(i);

				if (!a.getName().equals("SRC") &&
						!a.getName().equals("src") &&
						!a.getName().equals("style") &&
						!a.getName().equals("width") &&
						!a.getName().equals("height")) {
					toRemove.add(a);
				} // end if
			} // end for

			for (Iterator it = toRemove.iterator(); it.hasNext();) {
				Attr a = (Attr) it.next();
				((Element) node).removeAttributeNode(a);
			} // end for

			if (!base_url.endsWith("/")) {
				base_url = base_url + "/";
			} // end if

			Element img = (Element) node;
			String src = img.getAttribute("src");
			if ("".equals(src))
				src = img.getAttribute("SRC");
			if ("".equals(src)) {
			} else
			if (src.indexOf("://") == -1) {
				if (src.startsWith("/")) {
					String tmp    = base_url;
					String scheme = base_url.substring(0,
							base_url.indexOf("://"));
					tmp = tmp.substring(tmp.indexOf("://") + 3);

					String host = tmp.substring(0, tmp.indexOf("/"));
					img.setAttribute("src", scheme + "://" + host +
						src);
				} // end if
				else {
					if (src.startsWith("../")) {
						String scheme = base_url.substring(0,
								base_url.indexOf("://"));
						String host = base_url.substring(base_url.indexOf(
									"://") + 3);

						if (host.indexOf("/") != -1) {
							host = host.substring(0, host.indexOf("/"));
						} // end if

						base_url = base_url.substring(0,
								base_url.lastIndexOf("/"));

						while (src.startsWith("../")) {
							src = src.substring(3);

							if (base_url.indexOf("/") != -1) {
								base_url = base_url.substring(0,
										base_url.lastIndexOf("/"));
							} // end if
						} // end while

						if (base_url.indexOf(host) == -1) {
							base_url = scheme + "://" + host + "/";
						} // end if
						else {
							base_url = base_url + "/";
						} // end else

						img.setAttribute("src", base_url + src);
					} // end if
					else {
						img.setAttribute("src", base_url + src);
					} // end else
				} // end else
			} // end if
		} // end else if

		if ((node.getNodeType() == Node.ELEMENT_NODE) &&
				isInBody(node)) {
			CSSStyleDeclaration decl = styler.getStyle((Element) node,null);

			if (decl != null) {
				StringBuffer result = new StringBuffer();

				for (int i = 0; i < decl.getLength(); i++) {
					String declName = decl.item(i);
					String value = decl.getPropertyValue(declName);
					declName     = declName.replace('\"', '\'');
					value		 = value.replace('\"', '\'');
					result.append(declName);
					result.append(": ");
					result.append(CPadawan.getInstance()
											  .escapeAttribute(value));
					result.append("; ");
				} // end for

				Element elem = (Element) node;

				if (result.toString().trim().length() > 0) {
					elem.setAttribute("style", result.toString().trim());
				} // end if
			} // end if
		} // end if

		if (node.getChildNodes().getLength() > 0) {
			List     list = new ArrayList(node.getChildNodes().getLength());
			NodeList nl = node.getChildNodes();
	
			for (int i = 0; i < nl.getLength(); i++) {
				list.add(nl.item(i));
			} // end for
	
			for (int i = 0; i < list.size(); i++) {
				validateForSwingHtmlComponent((Node) list.get(i), base_url, styler);
			} // end for
		}
	} // end validateForSwingHtmlComponent()

	/**
	 * This class is used to render Form components. Radio, Checkbox,
	 * Combobox, Textfield, Textarea, Button
	 *
	 * @author Quentin Anciaux
	 * @version v0.98
	 */
	public static class CFormView
		extends FormView {
		/**
		 * Creates a new CFormView object.
		 *
		 * @param elem Swing text element representing the  form
		 * 		  component
		 */
		public CFormView(final javax.swing.text.Element elem) {
			super(elem);
		} // end CFormView()

		/**
		 * Do nothing in here
		 *
		 * @param fg not used
		 * @param a not used
		 */
		public void paint(
			final Graphics fg,
			final Shape    a) {}

		/**
		 * Paint the component on the graphics object
		 *
		 * @param fg The graphics object to paint on
		 * @param a the area to paint into.
		 */
		public void realPaint(
			final Graphics fg,
			final Shape    a) {
			super.paint(fg, a);

			Graphics2D g = (Graphics2D) fg;
			Component  c = getComponent();

			if (c == null) {
				c = createComponent();
			} // end if

			HTMLDocument doc  = (HTMLDocument) getDocument();
			Font		 font = doc.getFont(getAttributes());
			g.setFont(font);

			FontMetrics fm   = g.getFontMetrics();
			Rectangle   clip = g.getClipBounds();
			g.setClip(a);

			if (c instanceof JTextField) {
				JTextField t = (JTextField) c;
				CInputTextFieldRenderer.render(t, a, fm, g);
			} // end if
			else if (c instanceof JButton) {
				JButton t = (JButton) c;
				CInputButtonRenderer.render(t, a, fm, g);
			} // end else if
			else if (c instanceof JComboBox) {
				JComboBox t = (JComboBox) c;
				CSelectComboBoxRenderer.render(t, a, fm, g);
			} // end else if
			else if (c instanceof JCheckBox) {
				JCheckBox t = (JCheckBox) c;
				CInputCheckBoxRenderer.render(t, a, fm, g);
			} // end else if
			else if (c instanceof JRadioButton) {
				JRadioButton t = (JRadioButton) c;
				CInputRadioRenderer.render(t, a, fm, g);
			} // end else if
			else if (c instanceof JScrollPane) {
				JScrollPane pane = (JScrollPane) c;

				if (pane.getViewport().getComponents().length > 0) {
					c = pane.getViewport().getComponent(0);

					if (c instanceof JTextArea) {
						JTextArea t = (JTextArea) c;
						CTextareaRenderer.render(t, a, fm, g);
					} // end if
					else if (c instanceof JList) {
						JList t = (JList) c;
						CSelectListBoxRenderer.render(t, a, fm, g);
					} // end else if
				} // end if
			} // end else if

			g.setClip(clip);
		} // end realPaint()
	} // end CFormView

	/**
	 * This class extends ImageView and prevent images to be  painted
	 * while painting root container, thus letting YaHP calculates
	 * correct page cutting.
	 *
	 * @author Quentin Anciaux
	 * @version v0.98
	 */
	public static class CImageView
		extends ImageView {
		/**
		 * Creates a new CImageView object.
		 *
		 * @param elem The html swing element representing  the image
		 */
		public CImageView(final javax.swing.text.Element elem) {
			super(elem);
		} // end CImageView()

		/**
		 * If error while loading, do not display anything.
		 *
		 * @return No default icon, return null
		 */
		public Icon getNoImageIcon() {
			return null;
		} // end getNoImageIcon()

		/**
		 * do not paint
		 *
		 * @param g Graphics object
		 * @param a allocation
		 */
		public void paint(
			final Graphics g,
			final Shape    a) {} // end paint()

		/**
		 * Real painting method
		 *
		 * @param g Graphics object
		 * @param a allocation
		 */
		public void realPaint(
			final Graphics g,
			final Shape    a) {
			super.paint(g, a);
		} // end realPaint()
	} // end CImageView

	/**
	 * This class extends InlineView and prevent text to be  painted
	 * while painting root container, thus letting YaHP calculates
	 * correct page cutting.
	 *
	 * @author Quentin Anciaux
	 * @version v0.98
	 */
	public static class CInlineView
		extends javax.swing.text.html.InlineView {
		/**
		 * Creates a new CInlineView object.
		 *
		 * @param elem The html swing element representing  the text
		 * 		  node
		 */
		public CInlineView(final javax.swing.text.Element elem) {
			super(elem);
		} // end CInlineView()

		/**
		 * do not paint
		 *
		 * @param g Graphics object
		 * @param a allocation
		 */
		public void paint(
			final Graphics g,
			final Shape    a) {} // end paint()

		/**
		 * Real painting method
		 *
		 * @param gg Graphics object
		 * @param a allocation
		 */
		public void realPaint(
			final Graphics gg,
			final Shape    a) {
			Graphics2D g   = (Graphics2D) gg;
			Segment    s   = getText(getStartOffset(), getEndOffset());
			String     str = new String(s.array, s.offset, s.count).trim();

			if ((str.length() > 0) && isRightAlign(getElement())) {
				Rectangle2D r2d = g.getFontMetrics(getFont())
									   .getStringBounds(str, g);
				g.setFont(getFont());

				Rectangle r = a.getBounds();
				g.setColor(getForeground());

				int width = 0;

				for (int i = 0; i < str.length(); i++) {
					width += Math.round(g.getFontMetrics(getFont())
											 .getStringBounds("" +
								str.charAt(i), g).getWidth());
				} // end for

				int y	   = (int) Math.round((r.y + r2d.getHeight()) -
						g.getFontMetrics(getFont()).getDescent());
				int startx = (width < r.width)
					? (r.x + (r.width - width))
					: (r.x - (width - r.width));
				int x	   = startx;

				for (int i = 0; i < str.length(); i++) {
					g.drawString("" + str.charAt(i), x, y);
					x += Math.round(g.getFontMetrics(getFont())
										 .getStringBounds("" +
								str.charAt(i), g).getWidth());
				} // end for

				int starty = y;

				if (isStrikeThrough()) {
					y -= Math.round(r2d.getHeight() / 3);
					g.setStroke(new BasicStroke(1.0f));
					g.drawLine(startx, y, r.x + r.width, y);
				} // end if

				y = starty;

				if (isUnderline()) {
					g.setStroke(new BasicStroke(1.0f));
					g.drawLine(startx, y, r.x + r.width, y);
				} // end if
			} // end if
			else {
				super.paint(g, a);
			} // end else

		} // end realPaint()

		/**
		 * DOCUMENT ME!
		 *
		 * @param elem DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		private boolean isRightAlign(
			final javax.swing.text.Element elem) {
			if (elem == null) {
				return false;
			} // end if

			javax.swing.text.AttributeSet attrs = elem.getAttributes();

			for (Enumeration it = attrs.getAttributeNames();
					it.hasMoreElements();) {
				Object key = it.nextElement();

				if (key.toString().trim().equals("align")) {
					if ("right".equalsIgnoreCase(attrs.getAttribute(key)
														  .toString()
														  .trim())) {
						return true;
					} // end if

					return false;
				} // end if

				if (key.toString().trim().equals("text-align")) {
					if ("right".equalsIgnoreCase(attrs.getAttribute(key)
														  .toString()
														  .trim())) {
						return true;
					} // end if

					return false;
				} // end if
			} // end for

			return isRightAlign(elem.getParentElement());
		} // end isRightAlign()
	} // end CInlineView

	/**
	 * DOCUMENT ME!
	 *
	 * @author Quentin Anciaux
	 * @version 0.98
	 */
	public static class CView
		extends View {
		/** DOCUMENT ME! */
		View delegate;

		/**
		 * Creates a new CView object.
		 *
		 * @param delegate DOCUMENT ME!
		 * @param elem DOCUMENT ME!
		 */
		public CView(
			final View					   delegate,
			final javax.swing.text.Element elem) {
			super(elem);
			this.delegate = delegate;
		} // end CView()

		/**
		 * DOCUMENT ME!
		 *
		 * @param axis DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public float getAlignment(final int axis) {
			return delegate.getAlignment(axis);
		} // end getAlignment()

		/**
		 * DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public AttributeSet getAttributes() {
			return delegate.getAttributes();
		} // end getAttributes()

		/**
		 * DOCUMENT ME!
		 *
		 * @param axis DOCUMENT ME!
		 * @param pos DOCUMENT ME!
		 * @param len DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public int getBreakWeight(
			final int   axis,
			final float pos,
			final float len) {
			return delegate.getBreakWeight(axis, pos, len);
		} // end getBreakWeight()

		/**
		 * DOCUMENT ME!
		 *
		 * @param index DOCUMENT ME!
		 * @param a DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public Shape getChildAllocation(
			final int   index,
			final Shape a) {
			return delegate.getChildAllocation(index, a);
		} // end getChildAllocation()

		/**
		 * DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public Container getContainer() {
			return delegate.getContainer();
		} // end getContainer()

		/**
		 * DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public Class getDelegateClass() {
			return delegate.getClass();
		} // end getDelegateClass()

		/**
		 * DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public javax.swing.text.Document getDocument() {
			return delegate.getDocument();
		} // end getDocument()

		/**
		 * DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public javax.swing.text.Element getElement() {
			return delegate.getElement();
		} // end getElement()

		/**
		 * DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public int getEndOffset() {
			return delegate.getEndOffset();
		} // end getEndOffset()

		/**
		 * DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public Graphics getGraphics() {
			return delegate.getGraphics();
		} // end getGraphics()

		private String getWidth() {
			Object owidth = getElement().getAttributes().getAttribute(CSS.getAttribute("width"));
			if (owidth != null) {
				if (getParent() != null) {
					Object owidthp = getElement().getAttributes().getAttribute(CSS.getAttribute("width"));
					if (owidthp != null && owidthp.toString().equals(owidth.toString()))
						return null;
					return owidth.toString();
				}
			}
			return null;
		}
		
		private String getHeight() {
			Object oheight = getElement().getAttributes().getAttribute(CSS.getAttribute("height"));
			if (oheight != null) {
				if (getParent() != null) {
					Object oheightp = getElement().getAttributes().getAttribute(CSS.getAttribute("height"));
					if (oheightp != null && oheightp.toString().equals(oheight.toString()))
						return null;
					return oheight.toString();
				}
			}
			return null;
		}

		/**
		 * DOCUMENT ME!
		 *
		 * @param axis DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public float getMaximumSpan(final int axis) {
			if (axis == View.X_AXIS) {
				String width = getWidth();
				if (width != null) {
					try{
						return Float.parseFloat(width.substring(0,width.length()-2));
					}
					catch (Exception ignore){}
				}
			} else {
				String height = getHeight();
				if (height != null) {
					try{
						return Float.parseFloat(height.substring(0,height.length()-2));
					}
					catch (Exception ignore){}
				}
			}
			return delegate.getMaximumSpan(axis);
		} // end getMaximumSpan()

		/**
		 * DOCUMENT ME!
		 *
		 * @param axis DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public float getMinimumSpan(final int axis) {
			if (axis == View.X_AXIS) {
				String width = getWidth();
				if (width != null) {
					try{
						return Float.parseFloat(width.substring(0,width.length()-2));
					}
					catch (Exception ignore){}
				}
			} else {
				String height = getHeight();
				if (height != null) {
					try{
						return Float.parseFloat(height.substring(0,height.length()-2));
					}
					catch (Exception ignore){}
				}
			}
			return delegate.getMinimumSpan(axis);
		} // end getMinimumSpan()

		/**
		 * DOCUMENT ME!
		 *
		 * @param pos DOCUMENT ME!
		 * @param b DOCUMENT ME!
		 * @param a DOCUMENT ME!
		 * @param direction DOCUMENT ME!
		 * @param biasRet DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 *
		 * @throws BadLocationException DOCUMENT ME!
		 */
		public int getNextVisualPositionFrom(
			final int   pos,
			final Bias  b,
			final Shape a,
			final int   direction,
			final Bias  biasRet[])
			throws BadLocationException {
			return delegate.getNextVisualPositionFrom(pos, b, a,
				direction, biasRet);
		} // end getNextVisualPositionFrom()

		/**
		 * DOCUMENT ME!
		 *
		 * @param parent DOCUMENT ME!
		 */
		public void setParent(final View parent) {
			delegate.setParent(parent);
		} // end setParent()

		/**
		 * DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public View getParent() {
			return delegate.getParent();
		} // end getParent()

		/**
		 * DOCUMENT ME!
		 *
		 * @param axis DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public float getPreferredSpan(final int axis) {
			if (axis == View.X_AXIS) {
				String width = getWidth();
				if (width != null) {
					try{
						return Float.parseFloat(width.substring(0,width.length()-2));
					}
					catch (Exception ignore){}
				}
			} else {
				String height = getHeight();
				if (height != null) {
					try{
						return Float.parseFloat(height.substring(0,height.length()-2));
					}
					catch (Exception ignore){}
				}
			}
			return delegate.getPreferredSpan(axis);
		} // end getPreferredSpan()

		/**
		 * DOCUMENT ME!
		 *
		 * @param axis DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public int getResizeWeight(final int axis) {
			return delegate.getResizeWeight(axis);
		} // end getResizeWeight()

		/**
		 * DOCUMENT ME!
		 *
		 * @param width DOCUMENT ME!
		 * @param height DOCUMENT ME!
		 */
		public void setSize(
			final float width,
			final float height) {
			delegate.setSize(width, height);
		} // end setSize()

		/**
		 * DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public int getStartOffset() {
			return delegate.getStartOffset();
		} // end getStartOffset()

		/**
		 * DOCUMENT ME!
		 *
		 * @param x DOCUMENT ME!
		 * @param y DOCUMENT ME!
		 * @param allocation DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public String getToolTipText(
			final float x,
			final float y,
			final Shape allocation) {
			return delegate.getToolTipText(x, y, allocation);
		} // end getToolTipText()

		/**
		 * DOCUMENT ME!
		 *
		 * @param n DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public View getView(final int n) {
			return delegate.getView(n);
		} // end getView()

		/**
		 * DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public int getViewCount() {
			return delegate.getViewCount();
		} // end getViewCount()

		/**
		 * DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public ViewFactory getViewFactory() {
			return delegate.getViewFactory();
		} // end getViewFactory()

		/**
		 * DOCUMENT ME!
		 *
		 * @param x DOCUMENT ME!
		 * @param y DOCUMENT ME!
		 * @param allocation DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public int getViewIndex(
			final float x,
			final float y,
			final Shape allocation) {
			return delegate.getViewIndex(x, y, allocation);
		} // end getViewIndex()

		/**
		 * DOCUMENT ME!
		 *
		 * @param pos DOCUMENT ME!
		 * @param b DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public int getViewIndex(
			final int  pos,
			final Bias b) {
			return delegate.getViewIndex(pos, b);
		} // end getViewIndex()

		/**
		 * DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public boolean isVisible() {
			return delegate.isVisible();
		} // end isVisible()

		/**
		 * DOCUMENT ME!
		 *
		 * @param v DOCUMENT ME!
		 */
		public void append(final View v) {
			delegate.append(v);
		} // end append()

		/**
		 * DOCUMENT ME!
		 *
		 * @param axis DOCUMENT ME!
		 * @param offset DOCUMENT ME!
		 * @param pos DOCUMENT ME!
		 * @param len DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public View breakView(
			final int   axis,
			final int   offset,
			final float pos,
			final float len) {
			return delegate.breakView(axis, offset, pos, len);
		} // end breakView()

		/**
		 * DOCUMENT ME!
		 *
		 * @param e DOCUMENT ME!
		 * @param a DOCUMENT ME!
		 * @param f DOCUMENT ME!
		 */
		public void changedUpdate(
			final DocumentEvent e,
			final Shape			a,
			final ViewFactory   f) {
			delegate.changedUpdate(e, a, f);
		} // end changedUpdate()

		/**
		 * DOCUMENT ME!
		 *
		 * @param p0 DOCUMENT ME!
		 * @param p1 DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public View createFragment(
			final int p0,
			final int p1) {
			return delegate.createFragment(p0, p1);
		} // end createFragment()

		/**
		 * DOCUMENT ME!
		 *
		 * @param obj DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public boolean equals(final Object obj) {
			return obj == this || delegate.equals(obj);
		} // end equals()

		/**
		 * DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public int hashCode() {
			return delegate.hashCode();
		} // end hashCode()

		/**
		 * DOCUMENT ME!
		 *
		 * @param offs DOCUMENT ME!
		 * @param v DOCUMENT ME!
		 */
		public void insert(
			final int  offs,
			final View v) {
			delegate.insert(offs, v);
		} // end insert()

		/**
		 * DOCUMENT ME!
		 *
		 * @param e DOCUMENT ME!
		 * @param a DOCUMENT ME!
		 * @param f DOCUMENT ME!
		 */
		public void insertUpdate(
			final DocumentEvent e,
			final Shape			a,
			final ViewFactory   f) {
			delegate.insertUpdate(e, a, f);
		} // end insertUpdate()

		/**
		 * DOCUMENT ME!
		 *
		 * @param p0 DOCUMENT ME!
		 * @param b0 DOCUMENT ME!
		 * @param p1 DOCUMENT ME!
		 * @param b1 DOCUMENT ME!
		 * @param a DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 *
		 * @throws BadLocationException DOCUMENT ME!
		 */
		public Shape modelToView(
			final int   p0,
			final Bias  b0,
			final int   p1,
			final Bias  b1,
			final Shape a)
			throws BadLocationException {
			return delegate.modelToView(p0, b0, p1, b1, a);
		} // end modelToView()

		/**
		 * DOCUMENT ME!
		 *
		 * @param pos DOCUMENT ME!
		 * @param a DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 *
		 * @throws BadLocationException DOCUMENT ME!
		 */
		public Shape modelToView(
			final int   pos,
			final Shape a)
			throws BadLocationException {
			return delegate.modelToView(pos, a, null);
		} // end modelToView()

		/**
		 * DOCUMENT ME!
		 *
		 * @param pos DOCUMENT ME!
		 * @param a DOCUMENT ME!
		 * @param b DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 *
		 * @throws BadLocationException DOCUMENT ME!
		 */
		public Shape modelToView(
			final int   pos,
			final Shape a,
			final Bias  b)
			throws BadLocationException {
			return delegate.modelToView(pos, a, b);
		} // end modelToView()

		/**
		 * DOCUMENT ME!
		 *
		 * @param g DOCUMENT ME!
		 * @param allocation DOCUMENT ME!
		 */
		public void paint(
			final Graphics g,
			final Shape    allocation) {}

		/**
		 * DOCUMENT ME!
		 *
		 * @param child DOCUMENT ME!
		 * @param width DOCUMENT ME!
		 * @param height DOCUMENT ME!
		 */
		public void preferenceChanged(
			final View    child,
			final boolean width,
			final boolean height) {
			delegate.preferenceChanged(child, width, height);
		} // end preferenceChanged()

		/**
		 * DOCUMENT ME!
		 *
		 * @param g DOCUMENT ME!
		 * @param allocation DOCUMENT ME!
		 */
		public void realPaint(
			final Graphics g,
			final Shape    allocation) {
			delegate.paint(g, allocation);
			if (delegate.getClass().getName().equals("javax.swing.text.html.TableView")) {
				for (int i=0;i<getViewCount();i++) {
					View v = getView(i);
					// get rows
					CSwingHtmlBorder border = CSwingHtmlBorder.getBorder(v,true);
					Rectangle c = g.getClip().getBounds();
					Rectangle r = delegate.getChildAllocation(i,allocation).getBounds();
					if (r.x >= c.x && r.y >= c.y && r.x+r.width <= c.x+c.width && r.y+r.height <= c.y+c.height) {
						if (border.getBottomBorderWidth() > 0) {
							g.setColor(border.getBottomBorderColor());
							if (border.getBottomBorderType() == CSwingHtmlBorder.dashed) {
								((Graphics2D)g).setStroke(new BasicStroke(border.getBottomBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{6f,3f,1f},0.0f));
							} else
							if (border.getBottomBorderType() == CSwingHtmlBorder.dotted) {
								((Graphics2D)g).setStroke(new BasicStroke(border.getBottomBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{1f,3f,1f},0.0f));
							} else {
								((Graphics2D)g).setStroke(new BasicStroke(border.getBottomBorderWidth()));
							}
							g.drawLine(r.x,r.y+r.height,r.x+r.width,r.y+r.height);
						}
						if (border.getTopBorderWidth() > 0) {
							g.setColor(border.getTopBorderColor());
							if (border.getTopBorderType() == CSwingHtmlBorder.dashed) {
								((Graphics2D)g).setStroke(new BasicStroke(border.getTopBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{6f,3f,1f},0.0f));
							} else
							if (border.getTopBorderType() == CSwingHtmlBorder.dotted) {
								((Graphics2D)g).setStroke(new BasicStroke(border.getTopBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{1f,3f,1f},0.0f));
							} else {
								((Graphics2D)g).setStroke(new BasicStroke(border.getTopBorderWidth()));
							}
							g.drawLine(r.x,r.y,r.x+r.width,r.y);
						}
						if (border.getLeftBorderWidth() > 0) {
							g.setColor(border.getLeftBorderColor());
							if (border.getLeftBorderType() == CSwingHtmlBorder.dashed) {
								((Graphics2D)g).setStroke(new BasicStroke(border.getLeftBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{6f,3f,1f},0.0f));
							} else
							if (border.getLeftBorderType() == CSwingHtmlBorder.dotted) {
								((Graphics2D)g).setStroke(new BasicStroke(border.getLeftBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{1f,3f,1f},0.0f));
							} else {
								((Graphics2D)g).setStroke(new BasicStroke(border.getLeftBorderWidth()));
							}
							g.drawLine(r.x,r.y,r.x,r.y+r.height);
						}
						if (border.getRightBorderWidth() > 0) {
							g.setColor(border.getRightBorderColor());
							if (border.getRightBorderType() == CSwingHtmlBorder.dashed) {
								((Graphics2D)g).setStroke(new BasicStroke(border.getRightBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{6f,3f,1f},0.0f));
							} else
							if (border.getRightBorderType() == CSwingHtmlBorder.dotted) {
								((Graphics2D)g).setStroke(new BasicStroke(border.getRightBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{1f,3f,1f},0.0f));
							} else {
								((Graphics2D)g).setStroke(new BasicStroke(border.getRightBorderWidth()));
							}
							g.drawLine(r.x+r.width,r.y,r.x+r.width,r.y+r.height);
						}
					}
					// get Cells
					for (int j=0;j<v.getViewCount();j++) {
						View cell = v.getView(j);
						border = CSwingHtmlBorder.getBorder(cell,true);
						c = g.getClip().getBounds();
						r = delegate.getChildAllocation(i,allocation).getBounds();
						r = v.getChildAllocation(j,r).getBounds();
						if (r.x >= c.x && r.y >= c.y && r.x+r.width <= c.x+c.width && r.y+r.height <= c.y+c.height) {
							if (border.getBottomBorderWidth() > 0) {
								g.setColor(border.getBottomBorderColor());
								if (border.getBottomBorderType() == CSwingHtmlBorder.dashed) {
									((Graphics2D)g).setStroke(new BasicStroke(border.getBottomBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{6f,3f,1f},0.0f));
								} else
								if (border.getBottomBorderType() == CSwingHtmlBorder.dotted) {
									((Graphics2D)g).setStroke(new BasicStroke(border.getBottomBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{1f,3f,1f},0.0f));
								} else {
									((Graphics2D)g).setStroke(new BasicStroke(border.getBottomBorderWidth()));
								}
								g.drawLine(r.x,r.y+r.height,r.x+r.width,r.y+r.height);
							}
							if (border.getTopBorderWidth() > 0) {
								g.setColor(border.getTopBorderColor());
								if (border.getTopBorderType() == CSwingHtmlBorder.dashed) {
									((Graphics2D)g).setStroke(new BasicStroke(border.getTopBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{6f,3f,1f},0.0f));
								} else
								if (border.getTopBorderType() == CSwingHtmlBorder.dotted) {
									((Graphics2D)g).setStroke(new BasicStroke(border.getTopBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{1f,3f,1f},0.0f));
								} else {
									((Graphics2D)g).setStroke(new BasicStroke(border.getTopBorderWidth()));
								}
								g.drawLine(r.x,r.y,r.x+r.width,r.y);
							}
							if (border.getLeftBorderWidth() > 0) {
								g.setColor(border.getLeftBorderColor());
								if (border.getLeftBorderType() == CSwingHtmlBorder.dashed) {
									((Graphics2D)g).setStroke(new BasicStroke(border.getLeftBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{6f,3f,1f},0.0f));
								} else
								if (border.getLeftBorderType() == CSwingHtmlBorder.dotted) {
									((Graphics2D)g).setStroke(new BasicStroke(border.getLeftBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{1f,3f,1f},0.0f));
								} else {
									((Graphics2D)g).setStroke(new BasicStroke(border.getLeftBorderWidth()));
								}
								g.drawLine(r.x,r.y,r.x,r.y+r.height);
							}
							if (border.getRightBorderWidth() > 0) {
								g.setColor(border.getRightBorderColor());
								if (border.getRightBorderType() == CSwingHtmlBorder.dashed) {
									((Graphics2D)g).setStroke(new BasicStroke(border.getRightBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{6f,3f,1f},0.0f));
								} else
								if (border.getRightBorderType() == CSwingHtmlBorder.dotted) {
									((Graphics2D)g).setStroke(new BasicStroke(border.getRightBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{1f,3f,1f},0.0f));
								} else {
									((Graphics2D)g).setStroke(new BasicStroke(border.getRightBorderWidth()));
								}
								g.drawLine(r.x+r.width,r.y,r.x+r.width,r.y+r.height);
							}
						}
					}
				}
			} else {
				View parent = getParent();
				while (parent != null) {
					if (parent instanceof CView) {
						CView cp = (CView)parent;
						if (cp.delegate.getClass().getName().indexOf("javax.swing.text.html.TableView") != -1) {
							return;
						}
					} else if (parent != null && parent.getClass().getName().indexOf("javax.swing.text.html.TableView") != -1) {
						return;
					}
					parent = parent.getParent();
				}
				CSwingHtmlBorder border = CSwingHtmlBorder.getBorder(this,false);
				Rectangle c = g.getClip().getBounds();
				Rectangle r = allocation.getBounds();
				if (r.x >= c.x && r.y >= c.y && r.x+r.width <= c.x+c.width && r.y+r.height <= c.y+c.height) {
					if (border.getBottomBorderWidth() > 0) {
						g.setColor(border.getBottomBorderColor());
						if (border.getBottomBorderType() == CSwingHtmlBorder.dashed) {
							((Graphics2D)g).setStroke(new BasicStroke(border.getBottomBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{6f,3f,1f},0.0f));
						} else
						if (border.getBottomBorderType() == CSwingHtmlBorder.dotted) {
							((Graphics2D)g).setStroke(new BasicStroke(border.getBottomBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{1f,3f,1f},0.0f));
						} else {
							((Graphics2D)g).setStroke(new BasicStroke(border.getBottomBorderWidth()));
						}
						g.drawLine(r.x,r.y+r.height,r.x+r.width,r.y+r.height);
					}
					if (border.getTopBorderWidth() > 0) {
						g.setColor(border.getTopBorderColor());
						if (border.getTopBorderType() == CSwingHtmlBorder.dashed) {
							((Graphics2D)g).setStroke(new BasicStroke(border.getTopBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{6f,3f,1f},0.0f));
						} else
						if (border.getTopBorderType() == CSwingHtmlBorder.dotted) {
							((Graphics2D)g).setStroke(new BasicStroke(border.getTopBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{1f,3f,1f},0.0f));
						} else {
							((Graphics2D)g).setStroke(new BasicStroke(border.getTopBorderWidth()));
						}
						g.drawLine(r.x,r.y,r.x+r.width,r.y);
					}
					if (border.getLeftBorderWidth() > 0) {
						g.setColor(border.getLeftBorderColor());
						if (border.getLeftBorderType() == CSwingHtmlBorder.dashed) {
							((Graphics2D)g).setStroke(new BasicStroke(border.getLeftBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{6f,3f,1f},0.0f));
						} else
						if (border.getLeftBorderType() == CSwingHtmlBorder.dotted) {
							((Graphics2D)g).setStroke(new BasicStroke(border.getLeftBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{1f,3f,1f},0.0f));
						} else {
							((Graphics2D)g).setStroke(new BasicStroke(border.getLeftBorderWidth()));
						}
						g.drawLine(r.x,r.y,r.x,r.y+r.height);
					}
					if (border.getRightBorderWidth() > 0) {
						g.setColor(border.getRightBorderColor());
						if (border.getRightBorderType() == CSwingHtmlBorder.dashed) {
							((Graphics2D)g).setStroke(new BasicStroke(border.getRightBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{6f,3f,1f},0.0f));
						} else
						if (border.getRightBorderType() == CSwingHtmlBorder.dotted) {
							((Graphics2D)g).setStroke(new BasicStroke(border.getRightBorderWidth(),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10f,new float[]{1f,3f,1f},0.0f));
						} else {
							((Graphics2D)g).setStroke(new BasicStroke(border.getRightBorderWidth()));
						}
						g.drawLine(r.x+r.width,r.y,r.x+r.width,r.y+r.height);
					}
				}
			}
		} // end realPaint()

		/**
		 * DOCUMENT ME!
		 *
		 * @param i DOCUMENT ME!
		 */
		public void remove(final int i) {
			delegate.remove(i);
		} // end remove()

		/**
		 * DOCUMENT ME!
		 */
		public void removeAll() {
			delegate.removeAll();
		} // end removeAll()

		/**
		 * DOCUMENT ME!
		 *
		 * @param e DOCUMENT ME!
		 * @param a DOCUMENT ME!
		 * @param f DOCUMENT ME!
		 */
		public void removeUpdate(
			final DocumentEvent e,
			final Shape			a,
			final ViewFactory   f) {
			delegate.removeUpdate(e, a, f);
		} // end removeUpdate()

		/**
		 * DOCUMENT ME!
		 *
		 * @param offset DOCUMENT ME!
		 * @param length DOCUMENT ME!
		 * @param views DOCUMENT ME!
		 */
		public void replace(
			final int  offset,
			final int  length,
			final View views[]) {
			delegate.replace(offset, length, views);
		} // end replace()

		/**
		 * DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public String toString() {
			return delegate.toString();
		} // end toString()

		/**
		 * DOCUMENT ME!
		 *
		 * @param x DOCUMENT ME!
		 * @param y DOCUMENT ME!
		 * @param a DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public int viewToModel(
			final float x,
			final float y,
			final Shape a) {
			return delegate.viewToModel(x, y, a, new Bias[0]);
		} // end viewToModel()

		/**
		 * DOCUMENT ME!
		 *
		 * @param x DOCUMENT ME!
		 * @param y DOCUMENT ME!
		 * @param a DOCUMENT ME!
		 * @param biasReturn DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public int viewToModel(
			final float x,
			final float y,
			final Shape a,
			final Bias  biasReturn[]) {
			return delegate.viewToModel(x, y, a, biasReturn);
		} // end viewToModel()
	} // end CView

	/**
	 * Render a button
	 *
	 * @author Quentin Anciaux
	 * @version 0.98
	 */
	private static class CInputButtonRenderer {
		/**
		 * Render the given button in the graphics object
		 *
		 * @param t the button to render
		 * @param a the area to paint
		 * @param fm the current font metrics
		 * @param g the graphics 2d object
		 */
		public static void render(
			final JButton	  t,
			final Shape		  a,
			final FontMetrics fm,
			final Graphics2D  g) {
			Rectangle loc = a.getBounds();

			if (t.getIcon() == null) {
				if (fm.getHeight() > loc.height) {
					loc = new Rectangle(loc.x,
							loc.y -
							((fm.getHeight() - loc.height) / 2),
							loc.width, fm.getHeight());
					g.setClip(loc);
				} // end if

				g.setColor(new Color(231, 231, 231));
				g.fill3DRect(loc.x, loc.y, loc.width, loc.height, true);
				g.setColor(Color.BLACK);
				g.setStroke(new BasicStroke(2.0f));
				g.draw3DRect(loc.x, loc.y, loc.width, loc.height, true);

				if (t.getText() != null) {
					String text = t.getText().trim();
					g.drawString(text,
						loc.x + 2 +
						fm.getStringBounds(new char [] { 'm' }, 0, 1, g)
							  .getBounds().width,
						loc.y + ((loc.height - fm.getHeight()) / 2) +
						fm.getAscent() + fm.getLeading());
				} // end if
			} // end if
			else {
				Icon icon = t.getIcon();
				t.getIcon()
					 .paintIcon(t, g,
						loc.x +
						((loc.width - icon.getIconWidth()) / 2),
						loc.y +
						((loc.height - icon.getIconHeight()) / 2));
			} // end else
		} // end render()
	} // end CInputButtonRenderer

	/**
	 * Render a checkbox
	 *
	 * @author Quentin Anciaux
	 * @version 0.98
	 */
	private static class CInputCheckBoxRenderer {
		/**
		 * Render the given checkbox in the graphics object
		 *
		 * @param t the checkbox to render
		 * @param a the area to paint
		 * @param fm the current font metrics
		 * @param g the graphics 2d object
		 */
		public static void render(
			final JCheckBox   t,
			final Shape		  a,
			final FontMetrics fm,
			final Graphics2D  g) {
			Rectangle loc = a.getBounds();

			if (t.isSelected()) {
				g.setColor(Color.BLACK);
			} // end if
			else {
				g.setColor(Color.WHITE);
			} // end else

			g.fillRoundRect(loc.x + 4, loc.y + 8, loc.width - 17,
				loc.height - 17, 2, 2);
			g.setColor(Color.BLACK);
			g.setStroke(new BasicStroke(2.0f));
			g.drawRoundRect(loc.x, loc.y + 4, loc.width - 10,
				loc.height - 10, 2, 2);
		} // end render()
	} // end CInputCheckBoxRenderer

	/**
	 * Render a radio button
	 *
	 * @author Quentin Anciaux
	 * @version 0.98
	 */
	private static class CInputRadioRenderer {
		/**
		 * Render the given radio button in the graphics object
		 *
		 * @param t the radio button to render
		 * @param a the area to paint
		 * @param fm the current font metrics
		 * @param g the graphics 2d object
		 */
		public static void render(
			final JRadioButton t,
			final Shape		   a,
			final FontMetrics  fm,
			final Graphics2D   g) {
			Rectangle loc = a.getBounds();

			if (t.isSelected()) {
				g.setColor(Color.BLACK);
			} // end if
			else {
				g.setColor(Color.WHITE);
			} // end else

			g.fillOval(loc.x + 4, loc.y + 8, loc.width - 17,
				loc.height - 17);
			g.setColor(Color.BLACK);
			g.setStroke(new BasicStroke(2.0f));
			g.drawOval(loc.x, loc.y + 4, loc.width - 10, loc.height -
				10);
		} // end render()
	} // end CInputRadioRenderer

	/**
	 * Render an input text field
	 *
	 * @author Quentin Anciaux
	 * @version 0.98
	 */
	private static class CInputTextFieldRenderer {
		/**
		 * Render the given input text field in the graphics object
		 *
		 * @param t the input text field to render
		 * @param a the area to paint
		 * @param fm the current font metrics
		 * @param g the graphics 2d object
		 */
		public static void render(
			final JTextField  t,
			final Shape		  a,
			final FontMetrics fm,
			final Graphics2D  g) {
			Rectangle loc = a.getBounds();

			if (fm.getHeight() > loc.height) {
				loc = new Rectangle(loc.x,
						loc.y - ((fm.getHeight() - loc.height) / 2),
						loc.width, fm.getHeight());
				g.setClip(loc);
			} // end if

			g.setColor(t.getBackground());
			g.fill3DRect(loc.x, loc.y, loc.width, loc.height, true);
			g.setColor(t.getForeground());
			g.setStroke(new BasicStroke(2.0f));
			g.draw3DRect(loc.x, loc.y, loc.width, loc.height, true);
			g.drawString(t.getText(), loc.x + 4,
				loc.y + ((loc.height - fm.getHeight()) / 2) +
				fm.getAscent() + fm.getLeading());
		} // end render()
	} // end CInputTextFieldRenderer

	/**
	 * Render a combobox
	 *
	 * @author Quentin Anciaux
	 * @version 0.98
	 */
	private static class CSelectComboBoxRenderer {
		/**
		 * Render the given combobox in the graphics object
		 *
		 * @param t the combobox to render
		 * @param a the area to paint
		 * @param fm the current font metrics
		 * @param g the graphics 2d object
		 */
		public static void render(
			final JComboBox   t,
			final Shape		  a,
			final FontMetrics fm,
			final Graphics2D  g) {
			Rectangle loc   = a.getBounds();
			Rectangle bound = loc;

			if ((t.getSelectedItem() != null) &&
					(t.getSelectedItem().toString() != null)) {
				bound = fm.getStringBounds(t.getSelectedItem().toString(),
						g).getBounds();
			} // end if

			g.setColor(new Color(222, 224, 229));

			int width = (bound.width < loc.width)
				? loc.width
				: bound.width;

			if (width > loc.width) {
				g.setClip(new Rectangle(loc.x, loc.y, width, loc.height));
			} // end if

			g.fillRect(loc.x, loc.y, width, loc.height);
			g.setColor(Color.WHITE);
			g.fillRect(loc.x + 4, loc.y + 6, width - 8, loc.height -
				12);
			g.setColor(Color.BLACK);
			g.setStroke(new BasicStroke(2.0f));
			g.drawRect(loc.x, loc.y, width, loc.height);

			if ((t.getSelectedItem() != null) &&
					(t.getSelectedItem().toString() != null)) {
				g.drawString(t.getSelectedItem().toString(), loc.x + 5,
					loc.y + ((loc.height - fm.getHeight()) / 2) +
					fm.getAscent() + fm.getLeading());
			} // end if

			g.fill(new Polygon(
					new int [] {
						(loc.x + width) - 7, (loc.x + width) - 2,
						(loc.x + width) - 12
					},
					new int [] {
						(loc.y + loc.height) - 8, loc.y + 8, loc.y + 8
					},
					3));
		} // end render()
	} // end CSelectComboBoxRenderer

	/**
	 * Render a list box
	 *
	 * @author Quentin Anciaux
	 * @version 0.98
	 */
	private static class CSelectListBoxRenderer {
		/**
		 * Render the given list box in the graphics object
		 *
		 * @param t the list box to render
		 * @param a the area to paint
		 * @param fm the current font metrics
		 * @param g the graphics 2d object
		 */
		public static void render(
			final JList		  t,
			final Shape		  a,
			final FontMetrics fm,
			final Graphics2D  g) {
			Rectangle loc = a.getBounds();
			g.setColor(Color.WHITE);
			g.fillRect(loc.x, loc.y, loc.width, loc.height);
			g.setColor(Color.BLACK);
			g.setStroke(new BasicStroke(2.0f));
			g.drawRect(loc.x, loc.y, loc.width, loc.height);
			t.getModel().getSize();

			int offset = fm.getHeight();

			for (int i = 0; i < t.getModel().getSize(); i++) {
				String value = t.getModel().getElementAt(i).toString();
				g.drawString(value, loc.x + 5, loc.y + offset);
				offset = offset + fm.getAscent() + fm.getDescent() +
					fm.getLeading();

				if (offset > loc.height) {
					break;
				} // end if
			} // end for
		} // end render()
	} // end CSelectListBoxRenderer

	/**
	 * Render a textarea
	 *
	 * @author Quentin Anciaux
	 * @version 0.98
	 */
	private static class CTextareaRenderer {
		/**
		 * Render the given textarea in the graphics object
		 *
		 * @param t the textarea to render
		 * @param a the area to paint
		 * @param fm the current font metrics
		 * @param g the graphics 2d object
		 */
		public static void render(
			final JTextArea   t,
			final Shape		  a,
			final FontMetrics fm,
			final Graphics2D  g) {
			Rectangle loc = a.getBounds();
			g.setColor(Color.WHITE);
			g.fillRect(loc.x, loc.y, loc.width, loc.height);
			g.setColor(Color.BLACK);
			g.setStroke(new BasicStroke(2.0f));
			g.drawRect(loc.x, loc.y, loc.width, loc.height);

			char		 array[] = t.getText().toCharArray();
			StringBuffer buffer = new StringBuffer();
			int			 offset = fm.getHeight();

			for (int i = 0; i < array.length; i++) {
				if (array[i] == '\n') {
					g.drawString(buffer.toString(), loc.x + 5,
						loc.y + offset);
					buffer     = new StringBuffer(array[i]);
					offset     = offset + fm.getAscent() +
						fm.getDescent() + fm.getLeading();

					if (offset > loc.height) {
						break;
					} // end if

					continue;
				} // end if

				if (array[i] < 32) {
					array[i] = 32;
				} // end if

				buffer.append(array[i]);

				Rectangle2D bound = fm.getStringBounds(buffer.toString(),
						g);

				if (bound.getBounds().width >= loc.width) {
					buffer.delete(buffer.length() - 1, buffer.length());
					g.drawString(buffer.toString(), loc.x + 5,
						loc.y + offset);
					buffer     = new StringBuffer(array[i]);
					offset     = offset + fm.getAscent() +
						fm.getDescent() + fm.getLeading();
				} // end if

				if (offset > loc.height) {
					break;
				} // end if
			} // end for

			if ((buffer.length() > 0) && (offset <= loc.height)) {
				g.drawString(buffer.toString(), loc.x + 5,
					loc.y + offset);
			} // end if
		} // end render()
	} // end CTextareaRenderer

	/**
	 * An html editor kit which load images synchronously.
	 *
	 * @author Quentin Anciaux
	 * @version v0.98
	 */
	private static class SynchronousEditorKit
		extends HTMLEditorKit {
		/** serial version */
		static final long serialVersionUID = 1L;

		/**
		 * Creates a new SynchronousEditorKit object.
		 */
		public SynchronousEditorKit() {} // end SynchronousEditorKit()

		/**
		 * Return the view factory
		 *
		 * @return the view factory
		 */
		public ViewFactory getViewFactory() {
			return new SynchronousViewFactory(super.getViewFactory());
		} // end getViewFactory()
	} // end SynchronousEditorKit

	/**
	 * A view factory for the synchronous editor kit.
	 *
	 * @author Quentin Anciaux
	 * @version v0.98
	 */
	private static class SynchronousViewFactory
		implements ViewFactory {
		/** wrapped view factory */
		private ViewFactory base;

		/**
		 * Creates a new SynchronousViewFactory object.
		 *
		 * @param base wrapped view factory
		 */
		public SynchronousViewFactory(final ViewFactory base) {
			this.base = base;
		} // end SynchronousViewFactory()

		/**
		 * Create the view corresponding to an html element
		 *
		 * @param elem the html element
		 *
		 * @return the corresponding view object.
		 */
		public View create(final javax.swing.text.Element elem) {
			View result = base.create(elem);
			if (result instanceof ImageView) {
				result = new CImageView(elem);
				((CImageView) result).setLoadsSynchronously(true);
			} // end if
			else if (result.getClass() == javax.swing.text.html.InlineView.class) {
				result = new CInlineView(elem);
			} // end else if
			else if (result.getClass() == FormView.class) {
				result = new CFormView(elem);
			} // end else if
			else {
				result = new CView(result, elem);
			} // end else

			return result;
		} // end create()
	} // end SynchronousViewFactory
} // end CWebRendererSwingHtmlRenderer
