/*
 * Copyright (C) 2007 by Quentin Anciaux
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	@author Quentin Anciaux
 */
package org.allcolor.yahp.cl.converter.firefox;
import org.allcolor.yahp.cl.converter.CDocumentReconstructor;
import org.allcolor.yahp.converter.IHtmlToPdfTransformer;
import org.allcolor.yahp.converter.IHtmlToPdfTransformer.CConvertException;
import org.allcolor.yahp.converter.IHtmlToPdfTransformer.PageSize;

import org.apache.log4j.Logger;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;

import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Map;


/**
 * This class send an HTML document to be rendered in PS by Mozilla
 * firefox and converted in PDF by ps2pdf.  Firefox must be patched
 * with the furnished yahp patch for this to work. Options for
 * header/footer &T : Title  &U : Document URL  &D : Date/Time  &P :
 * Page Number  &PT : Page Number "of" Page Total
 *
 * @author Quentin Anciaux
 * @version 0.94
 */
public final class CFirefoxHtmlRenderer {
	/** Use for logging */
	private static final Logger log = Logger.getLogger(CFirefoxHtmlRenderer.class.getName());

	/**
	 * Constructor for the CFirefoxHtmlRenderer object
	 */
	private CFirefoxHtmlRenderer() {} // end CFirefoxHtmlRenderer()

	/**
	 * Render the document in a pdf.
	 *
	 * @param size page size
	 * @param file document
	 * @param fbase_url base url of the document
	 * @param hf headers footers list
	 * @param properties transform properties
	 * @param fout outputstream to write into
	 *
	 * @throws CConvertException if an error occurs while rendering
	 */
	public static final void renderToPDF(
		final PageSize	   size,
		final String	   file,
		final String	   fbase_url,
		final List		   hf,
		final Map		   properties,
		final OutputStream fout)
		throws CConvertException {
		OutputStream out		  = fout;
		String		 base_url     = fbase_url;
		String		 headerLeft   = null;
		String		 headerCentre = null;
		String		 headerRight  = null;
		String		 footerLeft   = null;
		String		 footerCentre = null;
		String		 footerRight  = null;

		if (hf != null) {
			for (Iterator it = hf.iterator(); it.hasNext();) {
				IHtmlToPdfTransformer.CHeaderFooter elem = (IHtmlToPdfTransformer.CHeaderFooter) it.next();

				if ("header-left".equals(elem.getType())) {
					headerLeft = elem.getContent();
				} // end if
				else if ("header-centre".equals(elem.getType())) {
					headerCentre = elem.getContent();
				} // end else if
				else if ("header-right".equals(elem.getType())) {
					headerRight = elem.getContent();
				} // end else if
				else if ("footer-left".equals(elem.getType())) {
					footerLeft = elem.getContent();
				} // end else if
				else if ("footer-centre".equals(elem.getType())) {
					footerCentre = elem.getContent();
				} // end else if
				else if ("footer-right".equals(elem.getType())) {
					footerRight = elem.getContent();
				} // end else if
			} // end for
		} // end if

		base_url = base_url.substring(0, base_url.lastIndexOf("/"));

		final double			   dsize[]   = size.getCMSize();
		final double			   dmargin[] = size.getCMMargin();

		final CFirefoxHtmlRenderer g2r		 = new CFirefoxHtmlRenderer();
		File					   fin		 = g2r.saveDocument(file,
				base_url);
		File					   fPS		 = null;
		File					   fPDF		 = null;

		try {
			fPS = File.createTempFile("yahp", ".ps");
			fPS.delete();
			fPDF = File.createTempFile("yahp", ".pdf");
			fPDF.delete();
			log.info("Convert : " + fin.toURI() + " into " +
				fPDF.toURI());

			String scriptPath = (properties == null)
				? "fireprint"
				: (String) properties.get(IHtmlToPdfTransformer.FIREFOX_PRINT_SCRIPT_PATH);

			if (scriptPath == null) {
				scriptPath = "fireprint";
			} // end if

			log.info("executing " + scriptPath);

			try {
				long    t1 = System.currentTimeMillis();
				Process p = Runtime.getRuntime().exec(scriptPath,
						new String [] {
							"MOZ_PS_FILENAME=" + fPS.getAbsolutePath(),
							"MOZ_PDF_CONVERT_URL=file://" +
							fin.getAbsolutePath(),
							"MOZ_PDF_FILENAME=" +
							fPDF.getAbsolutePath(),
							"MOZ_MARGIN_TOP=" + (dmargin[3] * 0.39),
							"MOZ_MARGIN_LEFT=" + (dmargin[0] * 0.39),
							"MOZ_MARGIN_BOTTOM=" + (dmargin[2] * 0.39),
							"MOZ_MARGIN_RIGHT=" + (dmargin[1] * 0.39),
							"MOZ_PAPER_WIDTH=" + (dsize[0] * 10),
							"MOZ_PAPER_HEIGHT=" + (dsize[1] * 10),
							"MOZ_HEADER_LEFT=" +
							((headerLeft == null)
							? ""
							: headerLeft),
							"MOZ_HEADER_CENTRE=" +
							((headerCentre == null)
							? ""
							: headerCentre),
							"MOZ_HEADER_RIGHT=" +
							((headerRight == null)
							? ""
							: headerRight),
							"MOZ_FOOTER_LEFT=" +
							((footerLeft == null)
							? ""
							: footerLeft),
							"MOZ_FOOTER_CENTRE=" +
							((footerCentre == null)
							? ""
							: footerCentre),
							"MOZ_FOOTER_RIGHT=" +
							((footerRight == null)
							? ""
							: footerRight)
						});
				new CStreamReader(p.getInputStream()).start();
				new CStreamReader(p.getErrorStream()).start();

				try {
					p.waitFor();
				} // end try
				catch (final Exception ignore) {} // end catch

				long t2   = System.currentTimeMillis();

				int  exit = p.exitValue();

				log.info("after executing " + scriptPath +
					" exit code: " + exit + " time : " + (t2 - t1) +
					"ms.");

				if (exit != 0) {
					throw new CConvertException(
						"Script exit with value != 0 : " + exit,
						null);
				} // end if
			} // end try
			catch (final IOException e) {
				throw new CConvertException(
					"ERROR: An IOException occured while calling Firefox: " +
					e.getMessage(),
					e);
			} // end catch

			CDocumentReconstructor.reconstruct(Arrays.asList(
					new File [] { fPDF }), properties, out, base_url,
				"Firefox Renderer");
		} // end try
		catch (final IOException e) {
			throw new CConvertException(
				"ERROR: An IOException occured while calling Firefox: " +
				e.getMessage(),
				e);
		} // end catch
		finally {
			try {
				fPDF.delete();
			} // end try
			catch (final Exception ignore) {}

			try {
				fPS.delete();
			} // end try
			catch (final Exception ignore) {}

			try {
				fin.delete();
			} // end try
			catch (final Exception ignore) {}
		} // end finally
	} // end renderToPDF()

	/**
	 * Sets the document attribute
	 *
	 * @param sfile The new document value
	 * @param base_url The base uri value
	 *
	 * @return A file containing the document
	 *
	 * @throws CConvertException if an error occurs
	 */
	private final File saveDocument(
		final String sfile,
		final String base_url)
		throws CConvertException {
		String file = sfile;
		file = file + "\n<script type=\"text/javascript\">\n" +
			"window.onload = function() { window.print();}\n" +
			"</script>\n";

		File tmp = null;

		try {
			tmp = File.createTempFile("yahp", ".html");

			FileOutputStream fo = new FileOutputStream(tmp);
			Writer			 w = new BufferedWriter(new OutputStreamWriter(
						fo,
						"utf-8"));
			w.write(file.toCharArray());
			w.flush();
			w.close();

			return tmp;
		} // end try
		catch (final IOException e) {
			try {
				tmp.delete();
			} // end try
			catch (final Exception ignore) {}

			throw new CConvertException(
				"ERROR: an IOException occured while saving document content: " +
				e.getMessage(),
				e);
		} // end catch
	} // end saveDocument()

	/**
	 * Read a stream in a separate thread.
	 *
	 * @author Quentin Anciaux
	 * @version 0.1
	 */
	private static class CStreamReader
		extends Thread {
		/** Encapsulated reader */
		private final BufferedReader in;

		/**
		 * Creates a new CStreamReader object.
		 *
		 * @param in the stream to read
		 */
		private CStreamReader(final InputStream in) {
			this.in = new BufferedReader(new InputStreamReader(in));
		} // end CStreamReader()

		/**
		 * read the stream and print in debug mode.
		 */
		public void run() {
			try {
				String line = null;

				while ((line = in.readLine()) != null) {
					if (log.isDebugEnabled()) {
						log.debug(line);
					} // end if
				} // end while
			} // end try
			catch (final Exception ignore) {}
		} // end run()
	} // end CStreamReader
} // end CFirefoxHtmlRenderer
