/*
 * Copyright (C) 2007 by Quentin Anciaux
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	@author Quentin Anciaux
 */
package org.allcolor.yahp.cl.converter.oo;
import com.sun.star.beans.PropertyValue;
import com.sun.star.frame.XComponentLoader;
import com.sun.star.frame.XStorable;
import com.sun.star.lang.XComponent;
import com.sun.star.uno.UnoRuntime;

import org.apache.log4j.Logger;

import java.io.FileNotFoundException;
import java.io.IOException;

import java.net.URI;


/**
 * Convert an html doc to pdf with OOo
 *
 * @author Quentin Anciaux
 * @version v0.94
 */
public class COOoConverter {
	/** used for logging */
	private static final Logger log = Logger.getLogger(COOoConverter.class);

	/** OOo properties to set */
	private static final PropertyValue OUTPUT_PROP[] = new PropertyValue[1];

	/** OOo empty property */
	private static final PropertyValue NO_VAL_PROP[] = new PropertyValue[0];

	static {
		OUTPUT_PROP[0]			 = new PropertyValue();
		OUTPUT_PROP[0].Name		 = "FilterName";
		OUTPUT_PROP[0].Value     = "writer_pdf_Export";
	} 

	/** connection handle to ooo */
	private COOoConnection connection;

	/**
	 * Creates a new COOoConverter object.
	 *
	 * @param connection connection handle to ooo
	 */
	public COOoConverter(final COOoConnection connection) {
		this.connection = connection;
	} // end COOoConverter()

	/**
	 * Convert the input html doc into output pdf doc
	 *
	 * @param input the input document
	 * @param output the output document
	 *
	 * @throws IOException if the output document is not writable (or
	 * 		   the filter name is wrong)
	 * @throws RuntimeException if the input document is not found
	 */
	public void convert(
		final URI input,
		final URI output)
		throws IOException {
		try {
			synchronized (connection) {
				XComponentLoader desktop  = connection.getDesktop();
				XComponent		 document = desktop.loadComponentFromURL(input.toString()
																				  .replaceAll("file:/",
								"file:///"), "_blank", 0, NO_VAL_PROP);

				if (document == null) {
					throw new FileNotFoundException(
						"input document not found: " + input);
				} // end if

				try {
					XStorable storable = (XStorable) UnoRuntime.queryInterface(XStorable.class,
							document);
					storable.storeToURL(output.toString()
												  .replaceAll("file:/",
								"file:///"), OUTPUT_PROP);
				} // end try
				finally {
					document.dispose();
				} // end finally
			} // end synchronized
		} // end try
		catch (final FileNotFoundException fnfe) {
			log.error("conversion failed; input = \"" + input +
				"\"; output = \"" + output + "\"; " +
				fnfe.getLocalizedMessage());
			throw fnfe;
		} // end catch
		catch (final com.sun.star.io.IOException ioe) {
			log.error("conversion failed; input = \"" + input +
				"\"; output = \"" + output + "\"; " +
				ioe.getLocalizedMessage());
			throw new IOException(ioe.getMessage());
		} // end catch
		catch (final Exception e) {
			log.error("conversion failed; input = \"" + input +
				"\"; output = \"" + output + "\"; " +
				e.getLocalizedMessage());
			throw new RuntimeException("conversion failed", e);
		} // end catch
	} // end convert()
} // end COOoConverter
