/*
 * Copyright (C) 2007 by Quentin Anciaux
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	@author Quentin Anciaux
 */
package org.allcolor.yahp.converter;
import java.awt.Toolkit;
import java.io.InputStream;
import java.io.OutputStream;

import java.util.List;
import java.util.Map;


/**
 * Interface for an html to pdf transformer.
 *
 * @author Quentin Anciaux
 * @version 1.2.11
 */
public interface IHtmlToPdfTransformer {
	/** YaHP Converter version */
	public static final String VERSION = "YaHP Converter v1.2.11 §11052007";

	/**
	 * If you'll convert non-latin html document, you can set this
	 * property to true or embed fonts. It affects speed, but 
	 * rendering will be correct. (Use it for chinese,japanese,...)
	 */
	public static final String HANDLE_CHAR_AS_GLYPH = "yahp.HANDLE_CHAR_AS_GLYPH";

	/** Set this property to true to allow encryption of  the PDF */
	public static final String USE_PDF_ENCRYPTION = "yahp.USE_PDF_ENCRYPTION";

	/**
	 * set to true to sign a document  with a certificate
	 * PDF_SIGNING_PRIVATE_KEY_PASSWORD,
	 * PDF_SIGNING_KEYSTORE_PASSWORD, PDF_SIGNING_PRIVATE_KEY_FILE
	 * must be set. PDF_SIGNING_REASON, PDF_SIGNING_LOCATION,
	 * USE_PDF_SELF_SIGNING may be set
	 */
	public static final String USE_PDF_SIGNING = "yahp.USE_PDF_SIGNING";

	/**
	 * default to true. Use an hand made certificate. To generate one :
	 * keytool -genkey -keyalg RSA -alias myname -keypass password
	 * -keystore keystore.ks -dname "cn=Your Name, c=CC" CC == Country
	 * code
	 */
	public static final String USE_PDF_SELF_SIGNING = "yahp.USE_PDF_SELF_SIGNING";

	/** set it with the path to the file containing the private key */
	public static final String PDF_SIGNING_PRIVATE_KEY_FILE = "yahp.PDF_SIGNING_PRIVATE_KEY_FILE";

	/** set it with a reason for the signing */
	public static final String PDF_SIGNING_REASON = "yahp.PDF_SIGNING_REASON";

	/** set it with the location of the signing */
	public static final String PDF_SIGNING_LOCATION = "yahp.PDF_SIGNING_LOCATION";

	/** set it with the password to open the keystore */
	public static final String PDF_SIGNING_KEYSTORE_PASSWORD = "yahp.PDF_SIGNING_KEYSTORE_PASSWORD";

	/** set it with the password to get the private key */
	public static final String PDF_SIGNING_PRIVATE_KEY_PASSWORD = "yahp.PDF_SIGNING_PRIVATE_KEY_PASSWORD";

	/** Default PDF renderer class, use swing to render HTML. */
	public static final String DEFAULT_PDF_RENDERER = "org.allcolor.yahp.cl.converter.CHtmlToPdfSwingTransformer";

	/** PDF renderer class, use OpenOffice.org to render HTML. */
	public static final String OOO_PDF_RENDERER = "org.allcolor.yahp.cl.converter.CHtmlToPdfOOoTransformer";
    
	/** PDF renderer class, use Mozilla Firefox to render HTML. */
	public static final String FIREFOX_PDF_RENDERER = "org.allcolor.yahp.cl.converter.CHtmlToPdfFirefoxTransformer";

	/** Firefox 'fireprint' script path.. */
	public static final String FIREFOX_PRINT_SCRIPT_PATH = "yahp.FIREFOX_PRINT_SCRIPT_PATH";

	/** OpenOffice listener hostname, default is localhost */
	public static final String OOO_PDF_HOST = "yahp.OOO_PDF_HOST";

	/** OpenOffice listener port, default is 8100 */
	public static final String OOO_PDF_PORT = "yahp.OOO_PDF_PORT";

	/** Path to TTF font for embedding with FOP */
	public static final String FOP_TTF_FONT_PATH = "yahp.FOP_TTF_FONT_PATH";

	/**
	 * Set this property with the desired password of pdf encryption.
	 * USE_PDF_ENCRYPTION must be set to true for this property to
	 * have an effect.
	 */
	public static final String PDF_ENCRYPTION_PASSWORD = "yahp.PDF_ENCRYPTION_PASSWORD";

	/**
	 * The operation permitted when the document is opened with the
	 * user password. USE_PDF_ENCRYPTION must be set to true for this
	 * property to have an effect.
	 */
	public static final String PDF_ALLOW_PRINTING = "yahp.PDF_ALLOW_PRINTING";

	/**
	 * The operation permitted when the document is opened with the
	 * user password. USE_PDF_ENCRYPTION must be set to true for this
	 * property to have an effect.
	 */
	public static final String PDF_ALLOW_MODIFY_CONTENTS = "yahp.PDF_ALLOW_MODIFY_CONTENTS";

	/**
	 * The operation permitted when the document is opened with the
	 * user password. USE_PDF_ENCRYPTION must be set to true for this
	 * property to have an effect.
	 */
	public static final String PDF_ALLOW_COPY = "yahp.PDF_ALLOW_COPY";

	/**
	 * The operation permitted when the document is opened with the
	 * user password. USE_PDF_ENCRYPTION must be set to true for this
	 * property to have an effect.
	 */
	public static final String PDF_ALLOW_MODIFT_ANNOTATIONS = "yahp.PDF_ALLOW_MODIFT_ANNOTATIONS";

	/**
	 * The operation permitted when the document is opened with the
	 * user password. USE_PDF_ENCRYPTION must be set to true for this
	 * property to have an effect.
	 */
	public static final String PDF_ALLOW_FILLIN = "yahp.PDF_ALLOW_FILLIN";

	/**
	 * The operation permitted when the document is opened with the
	 * user password. USE_PDF_ENCRYPTION must be set to true for this
	 * property to have an effect.
	 */
	public static final String PDF_ALLOW_SCREEN_READERS = "yahp.PDF_ALLOW_SCREEN_READERS";

	/**
	 * The operation permitted when the document is opened with the
	 * user password. USE_PDF_ENCRYPTION must be set to true for this
	 * property to have an effect.
	 */
	public static final String PDF_ALLOW_ASSEMBLY = "yahp.PDF_ALLOW_ASSEMBLY";

	/**
	 * The operation permitted when the document is opened with the
	 * user password. USE_PDF_ENCRYPTION must be set to true for this
	 * property to have an effect.
	 */
	public static final String PDF_ALLOW_DEGRADED_PRINTING = "yahp.PDF_ALLOW_DEGRADED_PRINTING";

	/** Set this property to set the pdf creator metadata. */
	public static final String PDF_CREATOR = "yahp.PDF_CREATOR";

	/** Set this property to set the pdf author metadata. */
	public static final String PDF_AUTHOR = "yahp.PDF_AUTHOR";

	/** Set this property to set the pdf title metadata. */
	public static final String PDF_TITLE = "yahp.PDF_TITLE";

	/** Set this property to set the pdf producer metadata. */
	public static final String PDF_PRODUCER = "yahp.PDF_PRODUCER";

	/** name of a custom pdf renderer class. */
	public static final String PDF_RENDERER_CLASS = "yahp.PDF_RENDERER_CLASS";

	/** A14 portrait size */
	public static final PageSize A14P = new PageSize(0.65d, 0.925d, 0.125d); //0037 x 0026 pixels   (09.25 x 06.50 mm)

	/** A13 portrait size */
	public static final PageSize A13P = new PageSize(0.925d, 1.3d, 0.125d); //0052 x 0037 pixels   (13.00 x 09.25 mm)

	/** A12 portrait size */
	public static final PageSize A12P = new PageSize(1.3d, 1.85d, 0.25d); //0074 x 0052 pixels   (18.50 x 13.00 mm)

	/** A11 portrait size */
	public static final PageSize A11P = new PageSize(1.85d, 2.6d, 0.25d); //0104 x 0074 pixels   (26.00 x 18.50 mm)

	/** A10 portrait size */
	public static final PageSize A10P = new PageSize(2.6d, 3.7d, 0.5d); //0148 x 0104 pixels   (003.7 x 002.6 cm)

	/** A9 portrait size */
	public static final PageSize A9P = new PageSize(3.7d, 5.2d, 0.5d); //0208 x 0148 pixels   (005.2 x 003.7 cm)

	/** A8 portrait size */
	public static final PageSize A8P = new PageSize(5.2d, 7.4d, 0.5d); //0296 x 0208 pixels   (007.4 x 005.2 cm)

	/** A7 portrait size */
	public static final PageSize A7P = new PageSize(7.4d, 10.4d, 0.5d); //0416 x 0296 pixels   (010.4 x 007.4 cm)

	/** A6 portrait size */
	public static final PageSize A6P = new PageSize(10.4d, 14.8d, 1d); //0592 x 0416 pixels   (014.8 x 010.4 cm)

	/** A5 portrait size */
	public static final PageSize A5P = new PageSize(14.8d, 20.8d, 1d); //0832 x 0592 pixels   (020.8 x 014.8 cm)

	/** A4 portrait size */
	public static final PageSize A4P = new PageSize(20.8d, 29.6d, 1d); //1184 x 0832 pixels   (029.6 x 020.8 cm)   (diagonal 36.3 cm or 15 inch)

	/** A3 portrait size */
	public static final PageSize A3P = new PageSize(29.6d, 41.6d, 1d); //1664 x 1184 pixels   (041.6 x 029.6 cm)   (diagonal 51.4 cm or 20 inch)

	/** A2 portrait size */
	public static final PageSize A2P = new PageSize(41.6d, 59.2d, 1d); //2368 x 1664 pixels   (059.2 x 041.6 cm)   (diagonal 72.6 cm or 28 inch)

	/** A1 portrait size */
	public static final PageSize A1P = new PageSize(59.2d, 83.2d, 1d); //3328 x 2368 pixels   (083.2 x 059.2 cm)

	/** A0 portrait size */
	public static final PageSize A0P = new PageSize(83.2d, 118.4d, 1d); //4736 x 3328 pixels   (118.4 x 084.0 cm)

	/** A14 landscape size */
	public static final PageSize A14L = new PageSize(0.925d, 0.65d, 0.125d); //0037 x 0026 pixels   (09.25 x 06.50 mm)

	/** A13 landscape size */
	public static final PageSize A13L = new PageSize(1.3d, 0.925d, 0.125d); //0052 x 0037 pixels   (13.00 x 09.25 mm)

	/** A12 landscape size */
	public static final PageSize A12L = new PageSize(1.85d, 1.3d, 0.25d); //0074 x 0052 pixels   (18.50 x 13.00 mm)

	/** A11 landscape size */
	public static final PageSize A11L = new PageSize(2.6d, 1.85d, 0.25d); //0104 x 0074 pixels   (26.00 x 18.50 mm)

	/** A10 landscape size */
	public static final PageSize A10L = new PageSize(3.7d, 2.6d, 0.5d); //0148 x 0104 pixels   (003.7 x 002.6 cm)

	/** A9 landscape size */
	public static final PageSize A9L = new PageSize(5.2d, 3.7d, 0.5d); //0208 x 0148 pixels   (005.2 x 003.7 cm)

	/** A8 landscape size */
	public static final PageSize A8L = new PageSize(7.4d, 5.2d, 0.5d); //0296 x 0208 pixels   (007.4 x 005.2 cm)

	/** A7 landscape size */
	public static final PageSize A7L = new PageSize(10.4d, 7.4d, 0.5d); //0416 x 0296 pixels   (010.4 x 007.4 cm)

	/** A6 landscape size */
	public static final PageSize A6L = new PageSize(14.8d, 10.4d, 1d); //0592 x 0416 pixels   (014.8 x 010.4 cm)

	/** A5 landscape size */
	public static final PageSize A5L = new PageSize(20.8d, 14.8d, 1d); //0832 x 0592 pixels   (020.8 x 014.8 cm)

	/** A4 landscape size */
	public static final PageSize A4L = new PageSize(29.6d, 20.8d, 1d); //1184 x 0832 pixels   (029.6 x 020.8 cm)   (diagonal 36.3 cm or 15 inch)

	/** A3 landscape size */
	public static final PageSize A3L = new PageSize(41.6d, 29.6d, 1d); //1664 x 1184 pixels   (041.6 x 029.6 cm)   (diagonal 51.4 cm or 20 inch)

	/** A2 landscape size */
	public static final PageSize A2L = new PageSize(59.2d, 41.6d, 1d); //2368 x 1664 pixels   (059.2 x 041.6 cm)   (diagonal 72.6 cm or 28 inch)

	/** A1 landscape size */
	public static final PageSize A1L = new PageSize(83.2d, 59.2d, 1d); //3328 x 2368 pixels   (083.2 x 059.2 cm)

	/** A0 landscape size */
	public static final PageSize A0L = new PageSize(118.4d, 83.2d, 40); //4736 x 3328 pixels   (118.4 x 084.0 cm)

	/**
	 * Transform the html document in the inputstream to a pdf in the
	 * outputstream
	 *
	 * @param in html document stream
	 * @param urlForBase base url of the document
	 * @param size pdf document page size
	 * @param hf header-footer list
	 * @param properties pdf renderer properties list
	 * @param out out stream to the pdf file precondition: in != null
	 * 		  precondition: urlForBase != null precondition: size !=
	 * 		  null precondition: != null precondition: properties !=
	 * 		  null precondition: out != null
	 *
	 * @throws CConvertException if a conversion exception occurs
	 */
	public void transform(
		final InputStream  in,
		final String	   urlForBase,
		final PageSize	   size,
		final List		   hf,
		final Map		   properties,
		final OutputStream out)
		throws CConvertException;

	/**
	 * An exception indicating a conversion error.
	 *
	 * @author Quentin Anciaux
	 * @version 0.1
	 */
	public static class CConvertException
		extends Exception {
		/** serialVersionUID */
		static final long serialVersionUID = -6405771637273004347L;

		/**
		 * Creates a new CConvertException object.
		 *
		 * @param message Error message
		 * @param cause DOCUMENT ME!
		 */
		public CConvertException(
			final String    message,
			final Throwable cause) {
			super(message, cause);
		} // end CConvertException()
	} // end CConvertException

	/**
	 * Represent a Header or a Footer
	 *
	 * @author Quentin Anciaux
	 * @version 0.3
	 */
	public static class CHeaderFooter {
		/** header/footer constant */
		public static final String HEADER = "header";

		/** header/footer constant */
		public static final String FOOTER = "footer";

		/** header/footer constant */
		public static final String HEADER_LEFT = "header-left";

		/** header/footer constant */
		public static final String HEADER_CENTRE = "header-centre";

		/** header/footer constant */
		public static final String HEADER_RIGHT = "header-right";

		/** header/footer constant */
		public static final String FOOTER_LEFT = "footer-left";

		/** header/footer constant */
		public static final String FOOTER_CENTRE = "footer-centre";

		/** header/footer constant */
		public static final String FOOTER_RIGHT = "footer-right";

		/** header/footer constant */
		public static final String EVEN_PAGES = "even";

		/** header/footer constant */
		public static final String ODD_PAGES = "odd";

		/** header/footer constant */
		public static final String ALL_PAGES = "all";

		/** html content of the element */
		private String content;

		/** for : even, odd or all */
		private String sfor;

		/**
		 * type: header or footer -- Swing  type:
		 * header[-right,-centre,-left] or
		 * footer[-right,-centre,-left] -- Firefox
		 */
		private String type;

		/**
		 * create a new CHeaderFooter
		 *
		 * @param content html content of the element
		 * @param type header[-right,-centre,-left] or
		 * 		 	  footer[-right,-centre,-left] -- Firefox
		 */
		public CHeaderFooter(
			final String content,
			final String type) {
			this(content, HEADER, ALL_PAGES);
			this.type		 = type;
			if (!type.equals(HEADER_LEFT) &&
				!type.equals(HEADER_CENTRE) &&
				!type.equals(HEADER_RIGHT) &&
				!type.equals(FOOTER_LEFT) &&
				!type.equals(FOOTER_CENTRE) &&
				!type.equals(FOOTER_RIGHT)) {
					throw new IllegalArgumentException(
						"type must be : [header|footer]-[left|centre|right].");
				} // end if
		} // end CHeaderFooter()

		/**
		 * create a new CHeaderFooter
		 *
		 * @param content html content of the element
		 * @param type header or footer
		 * @param sfor even, odd or all
		 *
		 * @throws IllegalArgumentException if type or header have
		 * 		   invalid value
		 */
		public CHeaderFooter(
			final String content,
			final String type,
			final String sfor) {
			this.content     = content;
			this.type		 = type;
			if (!type.equals(HEADER) &&
				!type.equals(FOOTER)) {
					throw new IllegalArgumentException(
						"type must be : [header|footer].");
			} // end if

			if (!sfor.equals(ALL_PAGES) &&
				!sfor.equals(ODD_PAGES) &&
				!sfor.equals(EVEN_PAGES)) {
				throw new IllegalArgumentException(
					"for is either : all,even or odd.");
			} // end if

			this.sfor = sfor;
		} // end CHeaderFooter()

		/**
		 * html content of the element
		 *
		 * @return html content of the element
		 */
		public final String getContent() {
			return content;
		} // end getContent()

		/**
		 * even, odd or all
		 *
		 * @return even, odd or all
		 */
		public final String getSfor() {
			return sfor;
		} // end getSfor()

		/**
		 * type: header or footer -- Swing  type:
		 * header[-right,-centre,-left] or
		 * footer[-right,-centre,-left] -- Firefox
		 *
		 * @return header or footer type
		 */
		public final String getType() {
			return type;
		} // end getType()
	} // end CHeaderFooter

	/**
	 * A class to setup the size of a pdf page.
	 *
	 * @author Quentin Anciaux
	 * @version 0.1
	 */
	public static final class PageSize {
		/** 1 centimeter is screenresolution in DPI/1 inch (2.54 cm) pixels */
		private static final double _1CMPIXEL = ((double)Toolkit.getDefaultToolkit().getScreenResolution())/2.54d;

		/**
		 * index 0 page left margin in centimeter   index 1 page right
		 * margin in centimeter index 2 page bottom margin in
		 * centimeter index 3 page top margin in centimeter
		 */
		private double cmmargin[] = new double[4];

		/**
		 * index 0 page width in centimeter   index 1 page height in
		 * centimeter
		 */
		private double cmwh[] = new double[2];

		/**
		 * index 0 page left margin in pixel   index 1 page right
		 * margin in pixel index 2 page bottom margin in pixel index 3
		 * page top margin in pixel
		 */
		private int margin[] = new int[4];

		/** index 0 page width in pixel   index 1 page height in pixel */
		private int wh[] = new int[2];

		/** page bottom margin in centimeter */
		private double cmbmargin;

		/** page height in centimeter */
		private double cmheight;

		/** page left margin in centimeter */
		private double cmlmargin;

		/** page right margin in centimeter */
		private double cmrmargin;

		/** page top margin in centimeter */
		private double cmtmargin;

		/** page width in centimeter */
		private double cmwidth;

		/** page bottom margin in pixel */
		private int bmargin;

		/** page height in pixel */
		private int height;

		/** page left margin in pixel */
		private int lmargin;

		/** page right margin in pixel */
		private int rmargin;

		/** page top margin in pixel */
		private int tmargin;

		/** page width in pixel */
		private int width;

		/**
		 * Creates a new PageSize object.
		 *
		 * @param width page width in pixel
		 * @param height page height in pixel
		 */
		public PageSize(
			final int width,
			final int height) {
			this(width, height, 0, 0, 0, 0);
		} // end PageSize()

		/**
		 * Creates a new PageSize object.
		 *
		 * @param width page width in pixel
		 * @param height page height in pixel
		 * @param margin page margin in pixel
		 */
		public PageSize(
			final int width,
			final int height,
			final int margin) {
			this(width, height, margin, margin, margin, margin);
		} // end PageSize()

		/**
		 * Creates a new PageSize object.
		 *
		 * @param width page width in pixel
		 * @param height page height in pixel
		 * @param lmargin page left margin in pixel
		 * @param rmargin page right margin in pixel
		 * @param bmargin page bottom margin in pixel
		 * @param tmargin page top margin in pixel
		 *
		 * @throws AssertionError if margin or width or height invalid
		 */
		public PageSize(
			final int width,
			final int height,
			final int lmargin,
			final int rmargin,
			final int bmargin,
			final int tmargin) {
			if ((width < 0) ||
					((height < 0) || (lmargin < 0) || (rmargin < 0) ||
					(bmargin < 0) || (tmargin < 0))) {
				throw new AssertionError(
					"Margin / Width / height invalid, aborting. - lmargin: " +
					lmargin + " - rmargin: " + rmargin +
					" - bmargin: " + bmargin + " - tmargin: " +
					tmargin + " - w: " + width + " - h: " + height);
			} // end if

			this.width		   = width;
			this.height		   = height;
			this.cmwidth	   = width / _1CMPIXEL;
			this.cmheight	   = height / _1CMPIXEL;
			this.lmargin	   = lmargin;
			this.rmargin	   = rmargin;
			this.bmargin	   = bmargin;
			this.tmargin	   = tmargin;
			this.cmlmargin     = lmargin / _1CMPIXEL;
			this.cmrmargin     = rmargin / _1CMPIXEL;
			this.cmbmargin     = bmargin / _1CMPIXEL;
			this.cmtmargin     = tmargin / _1CMPIXEL;
		} // end PageSize()

		/**
		 * Creates a new PageSize object.
		 *
		 * @param width page width in centimeter
		 * @param height page height in centimeter
		 */
		public PageSize(
			final double width,
			final double height) {
			this(width, height, 0, 0, 0, 0);
		} // end PageSize()

		/**
		 * Creates a new PageSize object.
		 *
		 * @param width page width in centimeter
		 * @param height page height in centimeter
		 * @param margin page margin in centimeter
		 */
		public PageSize(
			final double width,
			final double height,
			final double margin) {
			this(width, height, margin, margin, margin, margin);
		} // end PageSize()

		/**
		 * Creates a new PageSize object.
		 *
		 * @param width page width in centimeter
		 * @param height page height in centimeter
		 * @param lmargin page left margin in centimeter
		 * @param rmargin page right margin in centimeter
		 * @param bmargin page bottom margin in centimeter
		 * @param tmargin page top margin in centimeter
		 */
		public PageSize(
			final double width,
			final double height,
			final double lmargin,
			final double rmargin,
			final double bmargin,
			final double tmargin) {
			this((int) Math.round(_1CMPIXEL * width),
				(int) Math.round(_1CMPIXEL * height),
				(int) Math.round(_1CMPIXEL * lmargin),
				(int) Math.round(_1CMPIXEL * rmargin),
				(int) Math.round(_1CMPIXEL * bmargin),
				(int) Math.round(_1CMPIXEL * tmargin));
		} // end PageSize()

		/**
		 * index 0 page left margin in centimeter index 1 page right
		 * margin in centimeter index 2 page bottom margin in
		 * centimeter index 3 page top margin in centimeter
		 *
		 * @return double array size 4
		 */
		public double [] getCMMargin() {
			cmmargin[0]     = cmlmargin;
			cmmargin[1]     = cmrmargin;
			cmmargin[2]     = cmbmargin;
			cmmargin[3]     = cmtmargin;

			return cmmargin;
		} // end getCMMargin()

		/**
		 * index 0 page width in centimeter index 1 page height in
		 * centimeter
		 *
		 * @return double array size 2
		 */
		public double [] getCMSize() {
			cmwh[0]     = cmwidth;
			cmwh[1]     = cmheight;

			return cmwh;
		} // end getCMSize()

		/**
		 * index 0 page left margin in pixel index 1 page right margin
		 * in pixel index 2 page bottom margin in pixel index 3 page
		 * top margin in pixel
		 *
		 * @return int array size 4
		 */
		public int [] getMargin() {
			margin[0]     = lmargin;
			margin[1]     = rmargin;
			margin[2]     = bmargin;
			margin[3]     = tmargin;

			return margin;
		} // end getMargin()

		/**
		 * index 0 page width in pixel index 1 page height in pixel
		 *
		 * @return int array size 2
		 */
		public int [] getSize() {
			wh[0]     = width;
			wh[1]     = height;

			return wh;
		} // end getSize()
	} // end PageSize
} // end IHtmlToPdfTransformer
