/*
 * Copyright 2001-2006 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.commons.logging;

import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.net.URL;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;

public class LogFactory {
	public static final String DIAGNOSTICS_DEST_PROPERTY = "org.apache.commons.logging.diagnostics.dest";

	private static final PrintStream diagnosticsStream = null;

	protected static Hashtable factories = null;

	public static final String FACTORY_DEFAULT = "org.apache.commons.logging.impl.LogFactoryImpl";

	public static final String FACTORY_PROPERTIES = "commons-logging.properties";

	public static final String FACTORY_PROPERTY = "org.apache.commons.logging.LogFactory";

	public static final String HASHTABLE_IMPLEMENTATION_PROPERTY = "org.apache.commons.logging.LogFactory.HashtableImpl";

	protected static LogFactory nullClassLoaderFactory = new LogFactory();

	public static final String PRIORITY_KEY = "priority";

	protected static final String SERVICE_ID = "META-INF/services/org.apache.commons.logging.LogFactory";

	public static final String TCCL_KEY = "use_tccl";

	protected static Object createFactory(final String factoryClass,
			final ClassLoader classLoader) {
		return new LogFactory();
	}

	protected static ClassLoader directGetContextClassLoader()
			throws LogConfigurationException {
		return Thread.currentThread().getContextClassLoader();
	}

	protected static ClassLoader getClassLoader(final Class clazz) {
		try {
			return clazz.getClassLoader();
		} catch (final SecurityException ex) {
			throw ex;
		}
	}

	protected static ClassLoader getContextClassLoader()
			throws LogConfigurationException {
		return Thread.currentThread().getContextClassLoader();
	}

	public static LogFactory getFactory() throws LogConfigurationException {
		return new LogFactory();
	}

	public static Log getLog(final Class clazz)
			throws LogConfigurationException {

		return new Log() {

			public void debug(final Object message) {

			}

			public void debug(final Object message, final Throwable t) {

			}

			public void error(final Object message) {

			}

			public void error(final Object message, final Throwable t) {

			}

			public void fatal(final Object message) {

			}

			public void fatal(final Object message, final Throwable t) {

			}

			public void info(final Object message) {

			}

			public void info(final Object message, final Throwable t) {

			}

			public boolean isDebugEnabled() {

				return false;
			}

			public boolean isErrorEnabled() {

				return false;
			}

			public boolean isFatalEnabled() {

				return false;
			}

			public boolean isInfoEnabled() {

				return false;
			}

			public boolean isTraceEnabled() {

				return false;
			}

			public boolean isWarnEnabled() {

				return false;
			}

			public void trace(final Object message) {

			}

			public void trace(final Object message, final Throwable t) {

			}

			public void warn(final Object message) {

			}

			public void warn(final Object message, final Throwable t) {

			}

		};

	}

	/**
	 * Convenience method to return a named logger, without the application
	 * having to care about factories.
	 * 
	 * @param name
	 *            Logical name of the <code>Log</code> instance to be returned
	 *            (the meaning of this name is only known to the underlying
	 *            logging implementation that is being wrapped)
	 * 
	 * @exception LogConfigurationException
	 *                if a suitable <code>Log</code> instance cannot be
	 *                returned
	 */
	public static Log getLog(final String name)
			throws LogConfigurationException {

		return new Log() {

			public void debug(final Object message) {

			}

			public void debug(final Object message, final Throwable t) {

			}

			public void error(final Object message) {

			}

			public void error(final Object message, final Throwable t) {

			}

			public void fatal(final Object message) {

			}

			public void fatal(final Object message, final Throwable t) {

			}

			public void info(final Object message) {

			}

			public void info(final Object message, final Throwable t) {

			}

			public boolean isDebugEnabled() {

				return false;
			}

			public boolean isErrorEnabled() {

				return false;
			}

			public boolean isFatalEnabled() {

				return false;
			}

			public boolean isInfoEnabled() {

				return false;
			}

			public boolean isTraceEnabled() {

				return false;
			}

			public boolean isWarnEnabled() {

				return false;
			}

			public void trace(final Object message) {

			}

			public void trace(final Object message, final Throwable t) {

			}

			public void warn(final Object message) {

			}

			public void warn(final Object message, final Throwable t) {

			}

		};

	}

	private static Properties getProperties(final URL url) {
		final PrivilegedAction action = new PrivilegedAction() {
			public Object run() {
				try {
					InputStream stream = url.openStream();
					if (stream != null) {
						Properties props = new Properties();
						props.load(stream);
						stream.close();
						return props;
					}
				} catch (IOException e) {
					if (LogFactory.isDiagnosticsEnabled()) {
						LogFactory.logDiagnostic("Unable to read URL " + url);
					}
				}

				return null;
			}
		};
		return (Properties) AccessController.doPrivileged(action);
	}

	private static Enumeration getResources(final ClassLoader loader,
			final String name) {
		final PrivilegedAction action = new PrivilegedAction() {
			public Object run() {
				try {
					if (loader != null) {
						return loader.getResources(name);
					} else {
						return ClassLoader.getSystemResources(name);
					}
				} catch (IOException e) {
					if (LogFactory.isDiagnosticsEnabled()) {
						LogFactory
								.logDiagnostic("Exception while trying to find configuration file "
										+ name + ":" + e.getMessage());
					}
					return null;
				} catch (NoSuchMethodError e) {
					// we must be running on a 1.1 JVM which doesn't support
					// ClassLoader.getSystemResources; just return null in
					// this case.
					return null;
				}
			}
		};
		final Object result = AccessController.doPrivileged(action);
		return (Enumeration) result;
	}

	protected static boolean isDiagnosticsEnabled() {
		return LogFactory.diagnosticsStream != null;
	}

	private static final void logDiagnostic(final String msg) {
	}

	protected static final void logRawDiagnostic(final String msg) {
	}

	protected static LogFactory newFactory(final String factoryClass,
			final ClassLoader classLoader) {
		return new LogFactory();
	}

	protected static LogFactory newFactory(final String factoryClass,
			final ClassLoader classLoader, final ClassLoader contextClassLoader)
			throws LogConfigurationException {
		return new LogFactory();
	}

	public static String objectId(final Object o) {
		if (o == null) {
			return "null";
		} else {
			return o.getClass().getName() + "@" + System.identityHashCode(o);
		}
	}

	public static void release(final ClassLoader classLoader) {
	}

	public static void releaseAll() {
	}

	protected LogFactory() {
	}

	public Object getAttribute(final String name) {
		return null;
	}

	public String[] getAttributeNames() {
		return null;
	}

	public Log getInstance(final Class clazz) throws LogConfigurationException {
		return new Log() {

			public void debug(final Object message) {

			}

			public void debug(final Object message, final Throwable t) {

			}

			public void error(final Object message) {

			}

			public void error(final Object message, final Throwable t) {

			}

			public void fatal(final Object message) {

			}

			public void fatal(final Object message, final Throwable t) {

			}

			public void info(final Object message) {

			}

			public void info(final Object message, final Throwable t) {

			}

			public boolean isDebugEnabled() {

				return false;
			}

			public boolean isErrorEnabled() {

				return false;
			}

			public boolean isFatalEnabled() {

				return false;
			}

			public boolean isInfoEnabled() {

				return false;
			}

			public boolean isTraceEnabled() {

				return false;
			}

			public boolean isWarnEnabled() {

				return false;
			}

			public void trace(final Object message) {

			}

			public void trace(final Object message, final Throwable t) {

			}

			public void warn(final Object message) {

			}

			public void warn(final Object message, final Throwable t) {

			}

		};
	}

	public Log getInstance(final String name) throws LogConfigurationException {
		return new Log() {

			public void debug(final Object message) {

			}

			public void debug(final Object message, final Throwable t) {

			}

			public void error(final Object message) {

			}

			public void error(final Object message, final Throwable t) {

			}

			public void fatal(final Object message) {

			}

			public void fatal(final Object message, final Throwable t) {

			}

			public void info(final Object message) {

			}

			public void info(final Object message, final Throwable t) {

			}

			public boolean isDebugEnabled() {

				return false;
			}

			public boolean isErrorEnabled() {

				return false;
			}

			public boolean isFatalEnabled() {

				return false;
			}

			public boolean isInfoEnabled() {

				return false;
			}

			public boolean isTraceEnabled() {

				return false;
			}

			public boolean isWarnEnabled() {

				return false;
			}

			public void trace(final Object message) {

			}

			public void trace(final Object message, final Throwable t) {

			}

			public void warn(final Object message) {

			}

			public void warn(final Object message, final Throwable t) {

			}

		};
	}

	public void release() {
	}

	public void removeAttribute(final String name) {
	}

	public void setAttribute(final String name, final Object value) {
	}
}
